# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base class for farmOS master views
"""

import datetime
import json

import colander
import markdown
from webhelpers2.html import tags

from wuttaweb.views import MasterView
from wuttaweb.forms.schema import WuttaDateTime
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.util import get_farmos_client_for_user, use_farmos_style_grid_links
from wuttafarm.web.grids import (
    ResourceData,
    StringFilter,
    NullableStringFilter,
    DateTimeFilter,
    SimpleSorter,
)


class FarmOSMasterView(MasterView):
    """
    Base class for farmOS master views
    """

    model_key = "uuid"

    creatable = False
    editable = False
    deletable = False

    filterable = False
    sort_on_backend = False
    paginate_on_backend = False

    farmos_refurl_path = None

    labels = {
        "drupal_id": "Drupal ID",
        "raw_image_url": "Raw Image URL",
        "large_image_url": "Large Image URL",
        "thumbnail_image_url": "Thumbnail Image URL",
    }

    def __init__(self, request, context=None):
        super().__init__(request, context=context)
        self.farmos_client = get_farmos_client_for_user(self.request)
        self.farmos_4x = self.app.is_farmos_4x(self.farmos_client)
        self.normal = self.app.get_normalizer(self.farmos_client)
        self.raw_json = None
        self.farmos_style_grid_links = use_farmos_style_grid_links(self.config)

    def get_fallback_templates(self, template):
        """ """
        templates = super().get_fallback_templates(template)

        if template == "view":
            templates.insert(0, "/farmos/master/view.mako")

        return templates

    def render_owners_for_grid(self, obj, field, value):
        owners = []
        for user in value:
            if self.farmos_style_grid_links:
                url = self.request.route_url("farmos_users.view", uuid=user["uuid"])
                owners.append(tags.link_to(user["name"], url))
            else:
                owners.append(user["name"])
        return ", ".join(owners)

    def get_template_context(self, context):

        if self.listing and self.farmos_refurl_path:
            context["farmos_refurl"] = self.app.get_farmos_url(self.farmos_refurl_path)

        if self.viewing and self.raw_json:
            context["raw_json"] = self.raw_json
            code = "```json\n" + json.dumps(self.raw_json, indent=2) + "\n```"
            # TODO: this does not seem to be adding syntax highlight
            context["rendered_json"] = markdown.markdown(
                code, extensions=["fenced_code", "codehilite"]
            )

        return context


class TaxonomyMasterView(FarmOSMasterView):
    """
    Base class for farmOS "taxonomy term" views
    """

    farmos_taxonomy_type = None
    creatable = True
    editable = True
    deletable = True
    filterable = True
    sort_on_backend = True

    grid_columns = [
        "name",
        "description",
        "changed",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "changed",
    ]

    def get_grid_data(self, columns=None, session=None):
        return ResourceData(
            self.config,
            self.farmos_client,
            f"taxonomy_term--{self.farmos_taxonomy_type}",
            normalizer=self.normalize_taxonomy_term,
        )

    def normalize_taxonomy_term(self, term, included):

        if changed := term["attributes"]["changed"]:
            changed = datetime.datetime.fromisoformat(changed)
            changed = self.app.localtime(changed)

        if description := term["attributes"]["description"]:
            description = description["value"]

        return {
            "uuid": term["id"],
            "drupal_id": term["attributes"]["drupal_internal__tid"],
            "name": term["attributes"]["name"],
            "description": description or colander.null,
            "changed": changed,
        }

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")
        g.set_sorter("name", SimpleSorter("name"))
        g.set_filter("name", StringFilter)

        # description
        g.set_sorter("description", SimpleSorter("description.value"))
        g.set_filter("description", NullableStringFilter, path="description.value")

        # changed
        g.set_renderer("changed", "datetime")
        g.set_sorter("changed", SimpleSorter("changed"))
        g.set_filter("changed", DateTimeFilter)

    def get_instance(self):
        result = self.farmos_client.resource.get_id(
            "taxonomy_term", self.farmos_taxonomy_type, self.request.matchdict["uuid"]
        )
        self.raw_json = result
        return self.normalize_taxonomy_term(result["data"], {})

    def get_instance_title(self, term):
        return term["name"]

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # description
        f.set_widget("description", "notes")
        f.set_required("description", False)

        # changed
        if self.creating or self.editing:
            f.remove("changed")
        else:
            f.set_node("changed", WuttaDateTime())
            f.set_widget("changed", WuttaDateTimeWidget(self.request))

    def get_api_payload(self, term):

        attrs = {
            "name": term["name"],
        }

        if description := term["description"]:
            attrs["description"] = {"value": description}
        else:
            attrs["description"] = None

        return {"attributes": attrs}

    def persist(self, term, session=None):
        payload = self.get_api_payload(term)
        if self.editing:
            payload["id"] = term["uuid"]

        result = self.farmos_client.resource.send(
            "taxonomy_term", self.farmos_taxonomy_type, payload
        )

        if self.creating:
            term["uuid"] = result["data"]["id"]

    def delete_instance(self, term):
        self.farmos_client.resource.delete(
            "taxonomy_term", self.farmos_taxonomy_type, term["uuid"]
        )

    def get_xref_buttons(self, term):
        return [
            self.make_button(
                "View in farmOS",
                primary=True,
                url=self.app.get_farmos_url(f"/taxonomy/term/{term['drupal_id']}"),
                target="_blank",
                icon_left="external-link-alt",
            )
        ]



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
See also: :ref:`wuttafarm-import-farmos`
"""

import typer

from wuttasync.cli import import_command, ImportCommandHandler

from wuttafarm.cli import wuttafarm_typer


@wuttafarm_typer.command()
@import_command
def import_farmos(ctx: typer.Context, **kwargs):
    """
    Import data from farmOS API to WuttaFarm
    """
    config = ctx.parent.wutta_config
    handler = ImportCommandHandler(config, key="import.to_wuttafarm.from_farmos")
    handler.run(ctx)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
View for farmOS Seeding Logs
"""

from wuttaweb.forms.schema import WuttaDateTime
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views.farmos.logs import LogMasterView
from wuttafarm.web.grids import SimpleSorter, StringFilter


class SeedingLogView(LogMasterView):
    """
    View for farmOS seeding logs
    """

    model_name = "farmos_seeding_log"
    model_title = "farmOS Seeding Log"
    model_title_plural = "farmOS Seeding Logs"

    route_prefix = "farmos_logs_seeding"
    url_prefix = "/farmOS/logs/seeding"

    farmos_log_type = "seeding"
    farmos_refurl_path = "/logs/seeding"

    grid_columns = [
        "status",
        "drupal_id",
        "timestamp",
        "name",
        "assets",
        "locations",
        "purchase_date",
        "source",
        "is_group_assignment",
        "owners",
    ]

    def normalize_log(self, log, included):
        data = super().normalize_log(log, included)
        data.update(
            {
                "source": log["attributes"]["source"],
                "purchase_date": self.normal.normalize_datetime(
                    log["attributes"]["purchase_date"]
                ),
                "lot_number": log["attributes"]["lot_number"],
            }
        )
        return data

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # purchase_date
        g.set_renderer("purchase_date", "date")

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # source
        f.fields.insert_after("timestamp", "source")

        # purchase_date
        f.fields.insert_after("source", "purchase_date")
        f.set_node("purchase_date", WuttaDateTime())
        f.set_widget("purchase_date", WuttaDateTimeWidget(self.request))

        # lot_number
        f.fields.insert_after("purchase_date", "lot_number")


def defaults(config, **kwargs):
    base = globals()

    SeedingLogView = kwargs.get("SeedingLogView", base["SeedingLogView"])
    SeedingLogView.defaults(config)


def includeme(config):
    defaults(config)



"""add structure thumbnail url

Revision ID: 8cc1565d38e7
Revises: 2a49127e974b
Create Date: 2026-02-14 20:07:33.913573

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "8cc1565d38e7"
down_revision: Union[str, None] = "2a49127e974b"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # structure
    op.add_column(
        "structure", sa.Column("thumbnail_url", sa.String(length=255), nullable=True)
    )
    op.add_column(
        "structure_version",
        sa.Column(
            "thumbnail_url", sa.String(length=255), autoincrement=False, nullable=True
        ),
    )


def downgrade() -> None:

    # structure
    op.drop_column("structure_version", "thumbnail_url")
    op.drop_column("structure", "thumbnail_url")



"""add EquipmentAsset

Revision ID: e9b8664e1f39
Revises: e5b27eac471c
Create Date: 2026-03-09 18:05:54.917562

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "e9b8664e1f39"
down_revision: Union[str, None] = "e5b27eac471c"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # asset_equipment
    op.create_table(
        "asset_equipment",
        sa.Column("manufacturer", sa.String(length=255), nullable=True),
        sa.Column("model", sa.String(length=255), nullable=True),
        sa.Column("serial_number", sa.String(length=255), nullable=True),
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["uuid"], ["asset.uuid"], name=op.f("fk_asset_equipment_uuid_asset")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_asset_equipment")),
    )
    op.create_table(
        "asset_equipment_version",
        sa.Column(
            "manufacturer", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column("model", sa.String(length=255), autoincrement=False, nullable=True),
        sa.Column(
            "serial_number", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_asset_equipment_version")
        ),
    )
    op.create_index(
        op.f("ix_asset_equipment_version_end_transaction_id"),
        "asset_equipment_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_equipment_version_operation_type"),
        "asset_equipment_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_asset_equipment_version_pk_transaction_id",
        "asset_equipment_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_asset_equipment_version_pk_validity",
        "asset_equipment_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_equipment_version_transaction_id"),
        "asset_equipment_version",
        ["transaction_id"],
        unique=False,
    )

    # asset_equipment_equipment_type
    op.create_table(
        "asset_equipment_equipment_type",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("equipment_asset_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("equipment_type_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["equipment_asset_uuid"],
            ["asset_equipment.uuid"],
            name=op.f(
                "fk_asset_equipment_equipment_type_equipment_asset_uuid_asset_equipment"
            ),
        ),
        sa.ForeignKeyConstraint(
            ["equipment_type_uuid"],
            ["equipment_type.uuid"],
            name=op.f(
                "fk_asset_equipment_equipment_type_equipment_type_uuid_equipment_type"
            ),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_asset_equipment_equipment_type")),
    )
    op.create_table(
        "asset_equipment_equipment_type_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "equipment_asset_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "equipment_type_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid",
            "transaction_id",
            name=op.f("pk_asset_equipment_equipment_type_version"),
        ),
    )
    op.create_index(
        op.f("ix_asset_equipment_equipment_type_version_end_transaction_id"),
        "asset_equipment_equipment_type_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_equipment_equipment_type_version_operation_type"),
        "asset_equipment_equipment_type_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_asset_equipment_equipment_type_version_pk_transaction_id",
        "asset_equipment_equipment_type_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_asset_equipment_equipment_type_version_pk_validity",
        "asset_equipment_equipment_type_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_equipment_equipment_type_version_transaction_id"),
        "asset_equipment_equipment_type_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # asset_equipment_equipment_type
    op.drop_index(
        op.f("ix_asset_equipment_equipment_type_version_transaction_id"),
        table_name="asset_equipment_equipment_type_version",
    )
    op.drop_index(
        "ix_asset_equipment_equipment_type_version_pk_validity",
        table_name="asset_equipment_equipment_type_version",
    )
    op.drop_index(
        "ix_asset_equipment_equipment_type_version_pk_transaction_id",
        table_name="asset_equipment_equipment_type_version",
    )
    op.drop_index(
        op.f("ix_asset_equipment_equipment_type_version_operation_type"),
        table_name="asset_equipment_equipment_type_version",
    )
    op.drop_index(
        op.f("ix_asset_equipment_equipment_type_version_end_transaction_id"),
        table_name="asset_equipment_equipment_type_version",
    )
    op.drop_table("asset_equipment_equipment_type_version")
    op.drop_table("asset_equipment_equipment_type")

    # asset_equipment
    op.drop_index(
        op.f("ix_asset_equipment_version_transaction_id"),
        table_name="asset_equipment_version",
    )
    op.drop_index(
        "ix_asset_equipment_version_pk_validity", table_name="asset_equipment_version"
    )
    op.drop_index(
        "ix_asset_equipment_version_pk_transaction_id",
        table_name="asset_equipment_version",
    )
    op.drop_index(
        op.f("ix_asset_equipment_version_operation_type"),
        table_name="asset_equipment_version",
    )
    op.drop_index(
        op.f("ix_asset_equipment_version_end_transaction_id"),
        table_name="asset_equipment_version",
    )
    op.drop_table("asset_equipment_version")
    op.drop_table("asset_equipment")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base class for Asset master views
"""

import colander
import requests
from webhelpers2.html import tags

from wuttafarm.web.views.farmos import FarmOSMasterView
from wuttafarm.web.forms.schema import FarmOSRefs, FarmOSAssetRefs, FarmOSLocationRefs
from wuttafarm.web.forms.widgets import ImageWidget
from wuttafarm.web.grids import (
    ResourceData,
    StringFilter,
    IntegerFilter,
    BooleanFilter,
    SimpleSorter,
)


class AssetMasterView(FarmOSMasterView):
    """
    Base class for Asset master views
    """

    farmos_asset_type = None
    creatable = True
    editable = True
    deletable = True
    filterable = True
    sort_on_backend = True

    labels = {
        "name": "Asset Name",
        "asset_type_name": "Asset Type",
        "locations": "Location",
        "groups": "Group Membership",
        "thumbnail_url": "Thumbnail URL",
        "image_url": "Image URL",
    }

    grid_columns = [
        "thumbnail",
        "drupal_id",
        "name",
        "owners",
        "locations",
        "archived",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
        "archived": {"active": True, "verb": "is_false"},
    }

    form_fields = [
        "name",
        "notes",
        "asset_type_name",
        "owners",
        "is_location",
        "is_fixed",
        "locations",
        "groups",
        "archived",
        "drupal_id",
        "thumbnail_url",
        "image_url",
        "thumbnail",
        "image",
    ]

    def get_grid_data(self, **kwargs):
        return ResourceData(
            self.config,
            self.farmos_client,
            f"asset--{self.farmos_asset_type}",
            include=",".join(self.get_farmos_api_includes()),
            normalizer=self.normalize_asset,
        )

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # thumbnail
        g.set_renderer("thumbnail", self.render_grid_thumbnail)
        g.set_label("thumbnail", "", column_only=True)
        g.set_centered("thumbnail")

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)
        g.set_sorter("drupal_id", SimpleSorter("drupal_internal__id"))
        g.set_filter("drupal_id", IntegerFilter, path="drupal_internal__id")

        # name
        g.set_link("name")
        g.set_sorter("name", SimpleSorter("name"))
        g.set_filter("name", StringFilter)

        # owners
        g.set_label("owners", "Owner")
        g.set_renderer("owners", self.render_owners_for_grid)

        # locations
        g.set_renderer("locations", self.render_locations_for_grid)

        # groups
        g.set_label("groups", "Group Membership")
        g.set_renderer("groups", self.render_assets_for_grid)

        # archived
        g.set_renderer("archived", "boolean")
        g.set_sorter("archived", SimpleSorter("archived"))
        g.set_filter("archived", BooleanFilter)

    def render_grid_thumbnail(self, obj, field, value):
        if url := obj.get("thumbnail_url"):
            return tags.image(url, f"thumbnail for {self.get_model_title()}")
        return None

    def render_assets_for_grid(self, log, field, value):
        if not value:
            return ""

        assets = []
        for asset in value:
            if self.farmos_style_grid_links:
                route = f"farmos_{asset['asset_type']}_assets.view"
                url = self.request.route_url(route, uuid=asset["uuid"])
                assets.append(tags.link_to(asset["name"], url))
            else:
                assets.append(asset["name"])
        return ", ".join(assets)

    def render_locations_for_grid(self, asset, field, value):
        locations = []
        for location in value:
            if self.farmos_style_grid_links:
                asset_type = location["type"].split("--")[1]
                route = f"farmos_{asset_type}_assets.view"
                url = self.request.route_url(route, uuid=location["uuid"])
                locations.append(tags.link_to(location["name"], url))
            else:
                locations.append(location["name"])
        return ", ".join(locations)

    def grid_row_class(self, asset, data, i):
        """ """
        if asset["archived"]:
            return "has-background-warning"
        return None

    def get_farmos_api_includes(self):
        return {"asset_type", "location", "group", "owner", "image"}

    def get_instance(self):
        try:
            result = self.farmos_client.asset.get_id(
                self.farmos_asset_type,
                self.request.matchdict["uuid"],
                params={"include": ",".join(self.get_farmos_api_includes())},
            )
        except requests.HTTPError as exc:
            if exc.response.status_code == 404:
                raise self.notfound()
            raise
        self.raw_json = result
        included = {obj["id"]: obj for obj in result.get("included", [])}
        return self.normalize_asset(result["data"], included)

    def get_instance_title(self, asset):
        return asset["name"]

    def normalize_asset(self, asset, included):

        if notes := asset["attributes"]["notes"]:
            notes = notes["value"]

        if self.farmos_4x:
            archived = asset["attributes"]["archived"]
        else:
            archived = asset["attributes"]["status"] == "archived"

        asset_type_object = {}
        asset_type_name = None
        owner_objects = []
        owner_names = []
        location_objects = []
        location_names = []
        group_objects = []
        thumbnail_url = None
        image_url = None
        if relationships := asset.get("relationships"):

            if asset_type := relationships.get("asset_type"):
                if asset_type := included.get(asset_type["data"]["id"]):
                    asset_type_object = {
                        "uuid": asset_type["id"],
                        "name": asset_type["attributes"]["label"],
                    }
                    asset_type_name = asset_type_object["name"]

            if owners := relationships.get("owner"):
                for user in owners["data"]:
                    if user := included.get(user["id"]):
                        user = {
                            "uuid": user["id"],
                            "name": user["attributes"]["name"],
                        }
                        owner_objects.append(user)
                        owner_names.append(user["name"])

            if locations := relationships.get("location"):
                for location in locations["data"]:
                    if location := included.get(location["id"]):
                        location = {
                            "uuid": location["id"],
                            "type": location["type"],
                            "name": location["attributes"]["name"],
                        }
                        location_objects.append(location)
                        location_names.append(location["name"])

            if groups := relationships.get("group"):
                for group in groups["data"]:
                    if group := included.get(group["id"]):
                        group = {
                            "uuid": group["id"],
                            "name": group["attributes"]["name"],
                            "asset_type": "group",
                        }
                        group_objects.append(group)

            if images := relationships.get("image"):
                for image in images["data"]:
                    if image := included.get(image["id"]):
                        thumbnail_url = image["attributes"]["image_style_uri"][
                            "thumbnail"
                        ]
                        image_url = image["attributes"]["image_style_uri"]["large"]

        return {
            "uuid": asset["id"],
            "drupal_id": asset["attributes"]["drupal_internal__id"],
            "name": asset["attributes"]["name"],
            "asset_type": asset_type_object,
            "asset_type_name": asset_type_name,
            "is_location": asset["attributes"]["is_location"],
            "is_fixed": asset["attributes"]["is_fixed"],
            "notes": notes or colander.null,
            "owners": owner_objects,
            "owner_names": owner_names,
            "locations": location_objects,
            "location_names": location_names,
            "groups": group_objects,
            "archived": archived,
            "thumbnail_url": thumbnail_url or colander.null,
            "image_url": image_url or colander.null,
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        asset = f.model_instance

        # asset_type_name
        if self.creating or self.editing:
            f.remove("asset_type_name")

        # locations
        if self.creating or self.editing:
            f.remove("locations")
        else:
            f.set_label("locations", "Current Location")
            f.set_node("locations", FarmOSLocationRefs(self.request))

        # groups
        if self.creating or self.editing:
            f.remove("groups")  # TODO
        else:
            f.set_node("groups", FarmOSAssetRefs(self.request))

        # owners
        if self.creating or self.editing:
            f.remove("owners")  # TODO
        else:
            f.set_node("owners", FarmOSRefs(self.request, "farmos_users"))

        # notes
        f.set_widget("notes", "notes")
        f.set_required("notes", False)

        # is_location
        f.set_node("is_location", colander.Boolean())

        # is_fixed
        f.set_node("is_fixed", colander.Boolean())

        # groups
        if self.creating or self.editing:
            f.remove("groups")  # TODO: add support for this?

        # archived
        f.set_node("archived", colander.Boolean())

        # thumbnail_url
        if self.creating or self.editing:
            f.remove("thumbnail_url")

        # image_url
        if self.creating or self.editing:
            f.remove("image_url")

        # thumbnail
        if self.creating or self.editing:
            f.remove("thumbnail")
        elif asset.get("thumbnail_url"):
            f.set_widget("thumbnail", ImageWidget("asset thumbnail"))
            f.set_default("thumbnail", asset["thumbnail_url"])

        # image
        if self.creating or self.editing:
            f.remove("image")
        elif asset.get("image_url"):
            f.set_widget("image", ImageWidget("asset image"))
            f.set_default("image", asset["image_url"])

    def persist(self, asset, session=None):
        payload = self.get_api_payload(asset)
        if self.editing:
            payload["id"] = asset["uuid"]

        result = self.farmos_client.asset.send(self.farmos_asset_type, payload)

        if self.creating:
            asset["uuid"] = result["data"]["id"]

    def get_api_payload(self, asset):

        attrs = {
            "name": asset["name"],
            "notes": {"value": asset["notes"] or None},
            "archived": asset["archived"],
        }

        if "is_location" in asset:
            attrs["is_location"] = asset["is_location"]

        if "is_fixed" in asset:
            attrs["is_fixed"] = asset["is_fixed"]

        return {"attributes": attrs}

    def delete_instance(self, asset):
        self.farmos_client.asset.delete(self.farmos_asset_type, asset["uuid"])

    def get_xref_buttons(self, asset):
        return [
            self.make_button(
                "View in farmOS",
                primary=True,
                url=self.app.get_farmos_url(f"/asset/{asset['drupal_id']}"),
                target="_blank",
                icon_left="external-link-alt",
            ),
        ]



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model definition for Logs
"""

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.associationproxy import association_proxy

from wuttjamaican.db import model


class LogType(model.Base):
    """
    Represents a "log type" from farmOS
    """

    __tablename__ = "log_type"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Log Type",
        "model_title_plural": "Log Types",
    }

    uuid = model.uuid_column()

    name = sa.Column(
        sa.String(length=100),
        nullable=False,
        unique=True,
        doc="""
        Name of the log type.
        """,
    )

    description = sa.Column(
        sa.String(length=255),
        nullable=True,
        doc="""
        Optional description for the log type.
        """,
    )

    farmos_uuid = sa.Column(
        model.UUID(),
        nullable=True,
        unique=True,
        doc="""
        UUID for the log type within farmOS.
        """,
    )

    drupal_id = sa.Column(
        sa.String(length=50),
        nullable=True,
        unique=True,
        doc="""
        Drupal internal ID for the log type.
        """,
    )

    def __str__(self):
        return self.name or ""


class Log(model.Base):
    """
    Represents a base log record from farmOS
    """

    __tablename__ = "log"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Log",
        "model_title_plural": "All Logs",
    }

    uuid = model.uuid_column()

    log_type = sa.Column(
        sa.String(length=100),
        sa.ForeignKey("log_type.drupal_id"),
        nullable=False,
    )

    message = sa.Column(
        sa.String(length=255),
        nullable=False,
        doc="""
        Message text for the log.
        """,
    )

    timestamp = sa.Column(
        sa.DateTime(),
        nullable=False,
        doc="""
        Date and time when the log event occurred / will occur.
        """,
    )

    is_movement = sa.Column(
        sa.Boolean(),
        nullable=True,
        doc="""
        Whether the log represents a movement to new location.
        """,
    )

    is_group_assignment = sa.Column(
        sa.Boolean(),
        nullable=True,
        doc="""
        Whether the log represents a group assignment.
        """,
    )

    status = sa.Column(
        sa.String(length=20),
        nullable=False,
        doc="""
        Current status of the log event.
        """,
    )

    notes = sa.Column(
        sa.Text(),
        nullable=True,
        doc="""
        Arbitrary notes for the log event.
        """,
    )

    quick = sa.Column(
        sa.String(length=20),
        nullable=True,
        doc="""
        Identifier of quick form used to create the log, if
        applicable.
        """,
    )

    farmos_uuid = sa.Column(
        model.UUID(),
        nullable=True,
        unique=True,
        doc="""
        UUID for the log within farmOS.
        """,
    )

    drupal_id = sa.Column(
        sa.Integer(),
        nullable=True,
        unique=True,
        doc="""
        Drupal internal ID for the log.
        """,
    )

    _assets = orm.relationship(
        "LogAsset",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
        back_populates="log",
    )

    assets = association_proxy(
        "_assets",
        "asset",
        creator=lambda asset: LogAsset(asset=asset),
    )

    _groups = orm.relationship(
        "LogGroup",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
        back_populates="log",
    )

    groups = association_proxy(
        "_groups",
        "asset",
        creator=lambda asset: LogGroup(asset=asset),
    )

    _locations = orm.relationship(
        "LogLocation",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
        back_populates="log",
    )

    locations = association_proxy(
        "_locations",
        "asset",
        creator=lambda asset: LogLocation(asset=asset),
    )

    _quantities = orm.relationship(
        "LogQuantity",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
        back_populates="log",
    )

    quantities = association_proxy(
        "_quantities",
        "quantity",
        creator=lambda quantity: LogQuantity(quantity=quantity),
    )

    _owners = orm.relationship(
        "LogOwner",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
        back_populates="log",
    )

    owners = association_proxy(
        "_owners",
        "user",
        creator=lambda user: LogOwner(user=user),
    )

    def __str__(self):
        return self.message or ""


class LogMixin:

    uuid = model.uuid_fk_column("log.uuid", nullable=False, primary_key=True)

    @declared_attr
    def log(cls):
        return orm.relationship(
            Log,
            single_parent=True,
            cascade="all, delete-orphan",
            cascade_backrefs=False,
        )

    def __str__(self):
        return self.message or ""


def add_log_proxies(subclass):
    Log.make_proxy(subclass, "log", "farmos_uuid")
    Log.make_proxy(subclass, "log", "drupal_id")
    Log.make_proxy(subclass, "log", "log_type")
    Log.make_proxy(subclass, "log", "message")
    Log.make_proxy(subclass, "log", "timestamp")
    Log.make_proxy(subclass, "log", "is_movement")
    Log.make_proxy(subclass, "log", "is_group_assignment")
    Log.make_proxy(subclass, "log", "status")
    Log.make_proxy(subclass, "log", "notes")
    Log.make_proxy(subclass, "log", "quick")
    Log.make_proxy(subclass, "log", "assets")
    Log.make_proxy(subclass, "log", "groups")
    Log.make_proxy(subclass, "log", "locations")
    Log.make_proxy(subclass, "log", "quantities")
    Log.make_proxy(subclass, "log", "owners")


class LogAsset(model.Base):
    """
    Represents a "log's asset relationship" from farmOS.
    """

    __tablename__ = "log_asset"
    __versioned__ = {}

    uuid = model.uuid_column()

    log_uuid = model.uuid_fk_column("log.uuid", nullable=False)
    log = orm.relationship(
        Log,
        foreign_keys=log_uuid,
        back_populates="_assets",
    )

    asset_uuid = model.uuid_fk_column("asset.uuid", nullable=False)
    asset = orm.relationship(
        "Asset",
        foreign_keys=asset_uuid,
    )


class LogGroup(model.Base):
    """
    Represents a "log's group relationship" from farmOS.
    """

    __tablename__ = "log_group"
    __versioned__ = {}

    uuid = model.uuid_column()

    log_uuid = model.uuid_fk_column("log.uuid", nullable=False)
    log = orm.relationship(
        Log,
        foreign_keys=log_uuid,
        back_populates="_groups",
    )

    asset_uuid = model.uuid_fk_column("asset.uuid", nullable=False)
    asset = orm.relationship(
        "Asset",
        foreign_keys=asset_uuid,
    )


class LogLocation(model.Base):
    """
    Represents a "log's location relationship" from farmOS.
    """

    __tablename__ = "log_location"
    __versioned__ = {}

    uuid = model.uuid_column()

    log_uuid = model.uuid_fk_column("log.uuid", nullable=False)
    log = orm.relationship(
        Log,
        foreign_keys=log_uuid,
        back_populates="_locations",
    )

    asset_uuid = model.uuid_fk_column("asset.uuid", nullable=False)
    asset = orm.relationship(
        "Asset",
        foreign_keys=asset_uuid,
    )


class LogQuantity(model.Base):
    """
    Represents a "log's quantity relationship" from farmOS.
    """

    __tablename__ = "log_quantity"
    __versioned__ = {}

    uuid = model.uuid_column()

    log_uuid = model.uuid_fk_column("log.uuid", nullable=False)
    log = orm.relationship(
        Log,
        foreign_keys=log_uuid,
        back_populates="_quantities",
    )

    quantity_uuid = model.uuid_fk_column("quantity.uuid", nullable=False)
    quantity = orm.relationship(
        "Quantity",
        foreign_keys=quantity_uuid,
        back_populates="_log",
    )


class LogOwner(model.Base):
    """
    Represents a "log's owner relationship" from farmOS.
    """

    __tablename__ = "log_owner"
    __versioned__ = {}

    uuid = model.uuid_column()

    log_uuid = model.uuid_fk_column("log.uuid", nullable=False)
    log = orm.relationship(
        Log,
        foreign_keys=log_uuid,
        back_populates="_owners",
    )

    user_uuid = model.uuid_fk_column("user.uuid", nullable=False)
    user = orm.relationship(
        model.User,
        foreign_keys=user_uuid,
    )



"""add SeedingLog

Revision ID: dca5b48a5562
Revises: e9b8664e1f39
Create Date: 2026-03-10 09:52:13.999777

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "dca5b48a5562"
down_revision: Union[str, None] = "e9b8664e1f39"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # log_seeding
    op.create_table(
        "log_seeding",
        sa.Column("source", sa.String(length=255), nullable=True),
        sa.Column("purchase_date", sa.DateTime(), nullable=True),
        sa.Column("lot_number", sa.String(length=255), nullable=True),
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["uuid"], ["log.uuid"], name=op.f("fk_log_seeding_uuid_log")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_log_seeding")),
    )
    op.create_table(
        "log_seeding_version",
        sa.Column("source", sa.String(length=255), autoincrement=False, nullable=True),
        sa.Column("purchase_date", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column(
            "lot_number", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_log_seeding_version")
        ),
    )
    op.create_index(
        op.f("ix_log_seeding_version_end_transaction_id"),
        "log_seeding_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_seeding_version_operation_type"),
        "log_seeding_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_log_seeding_version_pk_transaction_id",
        "log_seeding_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_log_seeding_version_pk_validity",
        "log_seeding_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_seeding_version_transaction_id"),
        "log_seeding_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # log_seeding
    op.drop_index(
        op.f("ix_log_seeding_version_transaction_id"), table_name="log_seeding_version"
    )
    op.drop_index(
        "ix_log_seeding_version_pk_validity", table_name="log_seeding_version"
    )
    op.drop_index(
        "ix_log_seeding_version_pk_transaction_id", table_name="log_seeding_version"
    )
    op.drop_index(
        op.f("ix_log_seeding_version_operation_type"), table_name="log_seeding_version"
    )
    op.drop_index(
        op.f("ix_log_seeding_version_end_transaction_id"),
        table_name="log_seeding_version",
    )
    op.drop_table("log_seeding_version")
    op.drop_table("log_seeding")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Quick Form for "Eggs"
"""

import json

import colander
from deform.widget import SelectWidget

from farmOS.subrequests import Action, Subrequest, SubrequestsBlueprint, Format

from wuttaweb.forms.schema import WuttaDateTime
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views.quick import QuickFormView


class EggsQuickForm(QuickFormView):
    """
    Use this form to record an egg harvest. A harvest log will be
    created with standard details filled in.
    """

    form_title = "Eggs"
    route_slug = "eggs"
    url_slug = "eggs"

    _layer_assets = None

    # TODO: make this configurable?
    unit_name = "egg(s)"

    def make_quick_form(self):
        f = self.make_form(
            fields=[
                "timestamp",
                "count",
                "asset",
                "notes",
            ],
            labels={
                "timestamp": "Date",
                "count": "Quantity",
                "asset": "Layer Asset",
            },
        )

        # timestamp
        f.set_node("timestamp", WuttaDateTime())
        f.set_widget("timestamp", WuttaDateTimeWidget(self.request))
        f.set_default("timestamp", self.app.make_utc())

        # count
        f.set_node("count", colander.Integer())

        # asset
        assets = self.get_layer_assets()
        values = [(a["uuid"], a["name"]) for a in assets]
        f.set_widget("asset", SelectWidget(values=values))
        if len(assets) == 1:
            f.set_default("asset", assets[0]["uuid"])

        # notes
        f.set_widget("notes", "notes")
        f.set_required("notes", False)

        return f

    def get_layer_assets(self):
        if self._layer_assets is not None:
            return self._layer_assets

        if self.app.is_farmos_wrapper():
            assets = self.get_layer_assets_from_farmos()
        else:
            assets = self.get_layer_assets_from_wuttafarm()

        assets.sort(key=lambda a: a["name"])
        self._layer_assets = assets
        return assets

    def get_layer_assets_from_wuttafarm(self):
        model = self.app.model
        session = self.Session()
        assets = []

        def normalize(asset):
            asset_type = asset.__wutta_hint__["farmos_asset_type"]
            return {
                "uuid": str(asset.farmos_uuid),
                "name": asset.asset_name,
                "type": f"asset--{asset_type}",
            }

        query = (
            session.query(model.AnimalAsset)
            .join(model.Asset)
            .filter(model.AnimalAsset.produces_eggs == True)
            .order_by(model.Asset.asset_name)
        )
        assets.extend([normalize(a) for a in query])

        query = (
            session.query(model.GroupAsset)
            .join(model.Asset)
            .filter(model.GroupAsset.produces_eggs == True)
            .order_by(model.Asset.asset_name)
        )
        assets.extend([normalize(a) for a in query])

        return assets

    def get_layer_assets_from_farmos(self):
        assets = []
        params = {
            "filter[produces_eggs]": 1,
            "sort": "name",
        }

        def normalize(asset):
            return {
                "uuid": asset["id"],
                "name": asset["attributes"]["name"],
                "type": asset["type"],
            }

        result = self.farmos_client.asset.get("animal", params=params)
        assets.extend([normalize(a) for a in result["data"]])

        result = self.farmos_client.asset.get("group", params=params)
        assets.extend([normalize(a) for a in result["data"]])

        return assets

    def save_quick_form(self, form):

        if self.app.is_farmos_wrapper():
            return self.save_to_farmos(form)

        return self.save_to_wuttafarm(form)

    def save_to_farmos(self, form):
        data = form.validated

        assets = self.get_layer_assets()
        assets = {a["uuid"]: a for a in assets}
        asset = assets[data["asset"]]

        # TODO: make this configurable?
        unit_name = self.unit_name

        unit = {"data": {"type": "taxonomy_term--unit"}}
        new_unit = None

        result = self.farmos_client.resource.get(
            "taxonomy_term",
            "unit",
            params={
                "filter[name]": unit_name,
            },
        )
        if result["data"]:
            unit["data"]["id"] = result["data"][0]["id"]
        else:
            payload = dict(unit)
            payload["data"]["attributes"] = {"name": unit_name}
            new_unit = Subrequest(
                action=Action.create,
                requestId="create-unit",
                endpoint="api/taxonomy_term/unit",
                body=payload,
            )
            unit["data"]["id"] = "{{create-unit.body@$.data.id}}"

        quantity = {
            "data": {
                "type": "quantity--standard",
                "attributes": {
                    "measure": "count",
                    "value": {
                        "numerator": data["count"],
                        "denominator": 1,
                    },
                },
                "relationships": {
                    "units": unit,
                },
            },
        }

        kw = {}
        if new_unit:
            kw["waitFor"] = ["create-unit"]
        new_quantity = Subrequest(
            action=Action.create,
            requestId="create-quantity",
            endpoint="api/quantity/standard",
            body=quantity,
            **kw,
        )

        notes = None
        if data["notes"]:
            notes = {"value": data["notes"]}

        log = {
            "data": {
                "type": "log--harvest",
                "attributes": {
                    "name": f"Collected {data['count']} {unit_name}",
                    "timestamp": self.app.localtime(data["timestamp"]).timestamp(),
                    "notes": notes,
                    "quick": ["eggs"],
                },
                "relationships": {
                    "asset": {
                        "data": [
                            {
                                "id": asset["uuid"],
                                "type": asset["type"],
                            },
                        ],
                    },
                    "quantity": {
                        "data": [
                            {
                                "id": "{{create-quantity.body@$.data.id}}",
                                "type": "quantity--standard",
                            },
                        ],
                    },
                },
            },
        }

        new_log = Subrequest(
            action=Action.create,
            requestId="create-log",
            waitFor=["create-quantity"],
            endpoint="api/log/harvest",
            body=log,
        )

        blueprints = [new_quantity, new_log]
        if new_unit:
            blueprints.insert(0, new_unit)
        blueprint = SubrequestsBlueprint.parse_obj(blueprints)
        response = self.farmos_client.subrequests.send(blueprint, format=Format.json)

        log = json.loads(response["create-log#body{0}"]["body"])

        if self.app.is_farmos_mirror():
            if new_unit:
                unit = json.loads(response["create-unit"]["body"])
                self.app.auto_sync_from_farmos(
                    unit["data"], "Unit", client=self.farmos_client
                )
            quantity = json.loads(response["create-quantity"]["body"])
            self.app.auto_sync_from_farmos(
                quantity["data"], "StandardQuantity", client=self.farmos_client
            )
            self.app.auto_sync_from_farmos(
                log["data"], "HarvestLog", client=self.farmos_client
            )

        return log

    def save_to_wuttafarm(self, form):
        model = self.app.model
        session = self.Session()
        data = form.validated

        asset = (
            session.query(model.Asset)
            .filter(model.Asset.farmos_uuid == data["asset"])
            .one()
        )

        # TODO: make this configurable?
        unit_name = self.unit_name

        new_unit = False
        unit = session.query(model.Unit).filter(model.Unit.name == unit_name).first()
        if not unit:
            unit = model.Unit(name=unit_name)
            session.add(unit)
            new_unit = True

        quantity = model.StandardQuantity(
            quantity_type_id="standard",
            measure_id="count",
            value_numerator=data["count"],
            value_denominator=1,
            units=unit,
        )
        session.add(quantity)

        log = model.HarvestLog(
            log_type="harvest",
            message=f"Collected {data['count']} {unit_name}",
            timestamp=self.app.make_utc(data["timestamp"]),
            notes=data["notes"] or None,
            quick="eggs",
            status="done",
        )
        session.add(log)
        log.assets.append(asset)
        log.quantities.append(quantity.quantity)
        log.owners.append(self.request.user)
        session.flush()

        if self.app.is_farmos_mirror():
            if new_unit:
                self.app.auto_sync_to_farmos(unit, client=self.farmos_client)
            self.app.auto_sync_to_farmos(quantity, client=self.farmos_client)
            self.app.auto_sync_to_farmos(log, client=self.farmos_client)

        return log

    def redirect_after_save(self, log):
        model = self.app.model

        if isinstance(log, model.HarvestLog):
            return self.redirect(
                self.request.route_url("logs_harvest.view", uuid=log.uuid)
            )

        return self.redirect(
            self.request.route_url("farmos_logs_harvest.view", uuid=log["data"]["id"])
        )


def includeme(config):
    EggsQuickForm.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data normalizer for WuttaFarm / farmOS
"""

import datetime

from wuttjamaican.app import GenericHandler


class Normalizer(GenericHandler):
    """
    Base class and default implementation for the global data
    normalizer.  This should be used for normalizing records from
    WuttaFarm and/or farmOS.

    The point here is to have a single place to put the normalization
    logic, and let it be another thing which can be customized via
    subclass.
    """

    _farmos_units = None
    _farmos_measures = None

    def __init__(self, config, farmos_client=None):
        super().__init__(config)
        self.farmos_client = farmos_client

    def get_farmos_measures(self):
        if self._farmos_measures:
            return self._farmos_measures

        measures = {}
        response = self.farmos_client.session.get(
            self.app.get_farmos_url("/api/quantity/standard/resource/schema")
        )
        response.raise_for_status()
        data = response.json()
        for measure in data["definitions"]["attributes"]["properties"]["measure"][
            "oneOf"
        ]:
            measures[measure["const"]] = measure["title"]

        self._farmos_measures = measures
        return self._farmos_measures

    def get_farmos_measure_name(self, measure_id):
        measures = self.get_farmos_measures()
        return measures[measure_id]

    def get_farmos_unit(self, uuid):
        units = self.get_farmos_units()
        return units[uuid]

    def get_farmos_units(self):
        if self._farmos_units:
            return self._farmos_units

        units = {}
        result = self.farmos_client.resource.get("taxonomy_term", "unit")
        for unit in result["data"]:
            units[unit["id"]] = unit

        self._farmos_units = units
        return self._farmos_units

    def normalize_datetime(self, value):
        if not value:
            return None
        value = datetime.datetime.fromisoformat(value)
        return self.app.localtime(value)

    def normalize_farmos_asset(self, asset, included={}):
        """ """

        if notes := asset["attributes"]["notes"]:
            notes = notes["value"]

        parent_objects = []
        parent_uuids = []
        owner_objects = []
        owner_uuids = []
        if relationships := asset.get("relationships"):

            if parents := relationships.get("parent"):
                for parent in parents["data"]:
                    parent_uuid = parent["id"]
                    parent_uuids.append(parent_uuid)
                    parent_object = {
                        "uuid": parent_uuid,
                        "type": parent["type"],
                        "asset_type": parent["type"].split("--")[1],
                    }
                    if parent := included.get(parent_uuid):
                        parent_object.update(
                            {
                                "name": parent["attributes"]["name"],
                            }
                        )
                    parent_objects.append(parent_object)

            if owners := relationships.get("owner"):
                for user in owners["data"]:
                    user_uuid = user["id"]
                    owner_uuids.append(user_uuid)
                    if user := included.get(user_uuid):
                        owner_objects.append(
                            {
                                "uuid": user["id"],
                                "name": user["attributes"]["name"],
                            }
                        )

        # if self.farmos_4x:
        #     archived = asset["attributes"]["archived"]
        # else:
        #     archived = asset["attributes"]["status"] == "archived"

        return {
            "uuid": asset["id"],
            "drupal_id": asset["attributes"]["drupal_internal__id"],
            "asset_name": asset["attributes"]["name"],
            "is_location": asset["attributes"]["is_location"],
            "is_fixed": asset["attributes"]["is_fixed"],
            "archived": asset["attributes"]["archived"],
            "notes": notes,
            # nb. this is only used for certain asset types
            "produces_eggs": asset["attributes"].get("produces_eggs"),
            "parents": parent_objects,
            "parent_uuids": parent_uuids,
            "owners": owner_objects,
            "owner_uuids": owner_uuids,
        }

    def normalize_farmos_log(self, log, included={}):

        if timestamp := log["attributes"]["timestamp"]:
            timestamp = self.normalize_datetime(timestamp)

        if notes := log["attributes"]["notes"]:
            notes = notes["value"]

        log_type_object = {}
        log_type_uuid = None
        asset_objects = []
        group_objects = []
        group_uuids = []
        quantity_objects = []
        quantity_uuids = []
        location_objects = []
        location_uuids = []
        owner_objects = []
        owner_uuids = []
        if relationships := log.get("relationships"):

            if log_type := relationships.get("log_type"):
                log_type_uuid = log_type["data"]["id"]
                if log_type := included.get(log_type_uuid):
                    log_type_object = {
                        "uuid": log_type["id"],
                        "name": log_type["attributes"]["label"],
                    }

            if assets := relationships.get("asset"):
                for asset in assets["data"]:
                    asset_object = {
                        "uuid": asset["id"],
                        "type": asset["type"],
                        "asset_type": asset["type"].split("--")[1],
                    }
                    if asset := included.get(asset["id"]):
                        attrs = asset["attributes"]
                        rels = asset["relationships"]
                        asset_object.update(
                            {
                                "drupal_id": attrs["drupal_internal__id"],
                                "name": attrs["name"],
                                "is_location": attrs["is_location"],
                                "is_fixed": attrs["is_fixed"],
                                "archived": attrs["archived"],
                                "notes": attrs["notes"],
                            }
                        )
                    asset_objects.append(asset_object)

            if groups := relationships.get("group"):
                for group in groups["data"]:
                    group_uuid = group["id"]
                    group_uuids.append(group_uuid)
                    group_object = {
                        "uuid": group["id"],
                        "type": group["type"],
                        "asset_type": group["type"].split("--")[1],
                    }
                    if group := included.get(group_uuid):
                        attrs = group["attributes"]
                        rels = group["relationships"]
                        group_object.update(
                            {
                                "drupal_id": attrs["drupal_internal__id"],
                                "name": attrs["name"],
                                "is_location": attrs["is_location"],
                                "is_fixed": attrs["is_fixed"],
                                "archived": attrs["archived"],
                                "notes": attrs["notes"],
                            }
                        )
                    group_objects.append(group_object)

            if locations := relationships.get("location"):
                for location in locations["data"]:
                    location_uuid = location["id"]
                    location_uuids.append(location_uuid)
                    location_object = {
                        "uuid": location["id"],
                        "type": location["type"],
                        "asset_type": location["type"].split("--")[1],
                    }
                    if location := included.get(location_uuid):
                        attrs = location["attributes"]
                        rels = location["relationships"]
                        location_object.update(
                            {
                                "drupal_id": attrs["drupal_internal__id"],
                                "name": attrs["name"],
                                "is_location": attrs["is_location"],
                                "is_fixed": attrs["is_fixed"],
                                "archived": attrs["archived"],
                                "notes": attrs["notes"],
                            }
                        )
                    location_objects.append(location_object)

            if quantities := relationships.get("quantity"):
                for quantity in quantities["data"]:
                    quantity_uuid = quantity["id"]
                    quantity_uuids.append(quantity_uuid)
                    if quantity := included.get(quantity_uuid):
                        attrs = quantity["attributes"]
                        rels = quantity["relationships"]
                        value = attrs["value"]

                        unit_uuid = rels["units"]["data"]["id"]
                        unit = self.get_farmos_unit(unit_uuid)

                        measure_id = attrs["measure"]

                        quantity_object = {
                            "uuid": quantity["id"],
                            "drupal_id": attrs["drupal_internal__id"],
                            "quantity_type_uuid": rels["quantity_type"]["data"]["id"],
                            "quantity_type_id": rels["quantity_type"]["data"]["meta"][
                                "drupal_internal__target_id"
                            ],
                            "measure_id": measure_id,
                            "measure_name": self.get_farmos_measure_name(measure_id),
                            "value_numerator": value["numerator"],
                            "value_decimal": value["decimal"],
                            "value_denominator": value["denominator"],
                            "unit_uuid": unit_uuid,
                            "unit_name": unit["attributes"]["name"],
                        }
                        if quantity_object["quantity_type_id"] == "material":
                            quantity_object["material_types"] = [
                                {"uuid": mtype["id"]}
                                for mtype in quantity["relationships"]["material_type"][
                                    "data"
                                ]
                            ]
                        quantity_objects.append(quantity_object)

            if owners := relationships.get("owner"):
                for user in owners["data"]:
                    user_uuid = user["id"]
                    owner_uuids.append(user_uuid)
                    if user := included.get(user_uuid):
                        owner_objects.append(
                            {
                                "uuid": user["id"],
                                "name": user["attributes"]["name"],
                            }
                        )

        return {
            "uuid": log["id"],
            "drupal_id": log["attributes"]["drupal_internal__id"],
            "log_type_uuid": log_type_uuid,
            "log_type": log_type_object,
            "name": log["attributes"]["name"],
            "timestamp": timestamp,
            "assets": asset_objects,
            "groups": group_objects,
            "group_uuids": group_uuids,
            "quantities": quantity_objects,
            "quantity_uuids": quantity_uuids,
            "is_group_assignment": log["attributes"]["is_group_assignment"],
            "is_movement": log["attributes"]["is_movement"],
            "quick": log["attributes"]["quick"],
            "status": log["attributes"]["status"],
            "notes": notes,
            "locations": location_objects,
            "location_uuids": location_uuids,
            "owners": owner_objects,
            "owner_uuids": owner_uuids,
            # TODO: should we do this here or make caller do it?
            "vet": log["attributes"].get("vet"),
        }



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaFarm → farmOS data export
"""

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from wuttasync.importing import ImportHandler, FromWuttaHandler, FromWutta, Orientation

from wuttafarm.db import model
from wuttafarm.farmos import importing as farmos_importing


class FromWuttaFarmHandler(FromWuttaHandler):
    """
    Base class for import handler targeting WuttaFarm
    """

    source_key = "wuttafarm"


class ToFarmOSHandler(ImportHandler):
    """
    Base class for export handlers using CSV file(s) as data target.
    """

    target_key = "farmos"
    generic_target_title = "farmOS"

    # TODO: a lot of duplication to cleanup here; see FromFarmOSHandler

    def begin_target_transaction(self, client=None):
        """
        Establish the farmOS API client.
        """
        if client:
            self.farmos_client = client
        else:
            token = self.get_farmos_oauth2_token()
            self.farmos_client = self.app.get_farmos_client(token=token)
        self.farmos_4x = self.app.is_farmos_4x(self.farmos_client)

    def get_farmos_oauth2_token(self):

        client_id = self.config.get(
            "farmos.oauth2.importing.client_id", default="wuttafarm"
        )
        client_secret = self.config.require("farmos.oauth2.importing.client_secret")
        scope = self.config.get("farmos.oauth2.importing.scope", default="farm_manager")

        client = BackendApplicationClient(client_id=client_id)
        oauth = OAuth2Session(client=client)

        return oauth.fetch_token(
            token_url=self.app.get_farmos_url("/oauth/token"),
            include_client_id=True,
            client_secret=client_secret,
            scope=scope,
        )

    def get_importer_kwargs(self, key, **kwargs):
        kwargs = super().get_importer_kwargs(key, **kwargs)
        kwargs["farmos_client"] = self.farmos_client
        kwargs["farmos_4x"] = self.farmos_4x
        return kwargs


class FromWuttaFarmToFarmOS(FromWuttaFarmHandler, ToFarmOSHandler):
    """
    Handler for WuttaFarm → farmOS API export.
    """

    orientation = Orientation.EXPORT

    def define_importers(self):
        """ """
        importers = super().define_importers()
        importers["LandAsset"] = LandAssetImporter
        importers["StructureAsset"] = StructureAssetImporter
        importers["WaterAsset"] = WaterAssetImporter
        importers["EquipmentType"] = EquipmentTypeImporter
        importers["EquipmentAsset"] = EquipmentAssetImporter
        importers["AnimalType"] = AnimalTypeImporter
        importers["AnimalAsset"] = AnimalAssetImporter
        importers["GroupAsset"] = GroupAssetImporter
        importers["PlantType"] = PlantTypeImporter
        importers["Season"] = SeasonImporter
        importers["PlantAsset"] = PlantAssetImporter
        importers["Unit"] = UnitImporter
        importers["MaterialType"] = MaterialTypeImporter
        importers["MaterialQuantity"] = MaterialQuantityImporter
        importers["StandardQuantity"] = StandardQuantityImporter
        importers["ActivityLog"] = ActivityLogImporter
        importers["HarvestLog"] = HarvestLogImporter
        importers["MedicalLog"] = MedicalLogImporter
        importers["ObservationLog"] = ObservationLogImporter
        importers["SeedingLog"] = SeedingLogImporter
        return importers


class FromWuttaFarm(FromWutta):

    drupal_internal_id_field = "drupal_internal__id"

    def create_target_object(self, key, source_data):
        obj = super().create_target_object(key, source_data)
        if obj is None:
            return None

        if not self.dry_run:

            # set farmOS, Drupal key fields in WuttaFarm
            api_object = obj["_new_object"]
            wf_object = source_data["_src_object"]
            wf_object.farmos_uuid = obj["uuid"]
            wf_object.drupal_id = api_object["attributes"][
                self.drupal_internal_id_field
            ]

        return obj


class FromWuttaFarmAsset(FromWuttaFarm):
    """
    Base class for WuttaFarm → farmOS API asset exporters
    """

    supported_fields = [
        "uuid",
        "asset_name",
        "is_location",
        "is_fixed",
        "parents",
        "notes",
        "archived",
    ]

    def normalize_source_object(self, asset):
        return {
            "uuid": asset.farmos_uuid or self.app.make_true_uuid(),
            "asset_name": asset.asset_name,
            "is_location": asset.is_location,
            "is_fixed": asset.is_fixed,
            "parents": [(p.asset_type, p.farmos_uuid) for p in asset.parents],
            "notes": asset.notes,
            "archived": asset.archived,
            "_src_object": asset,
        }


class AnimalAssetImporter(
    FromWuttaFarmAsset, farmos_importing.model.AnimalAssetImporter
):
    """
    WuttaFarm → farmOS API exporter for Animal Assets
    """

    source_model_class = model.AnimalAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "animal_type_uuid",
                "sex",
                "is_sterile",
                "produces_eggs",
                "birthdate",
            ]
        )
        return fields

    def normalize_source_object(self, animal):
        data = super().normalize_source_object(animal)
        data.update(
            {
                "animal_type_uuid": animal.animal_type.farmos_uuid,
                "sex": animal.sex,
                "is_sterile": animal.is_sterile,
                "produces_eggs": animal.produces_eggs,
                "birthdate": animal.birthdate,
            }
        )
        return data


class FromWuttaFarmTaxonomy(FromWuttaFarm):
    """
    Base class for taxonomy term exporters
    """

    supported_fields = [
        "uuid",
        "name",
        "description",
    ]

    drupal_internal_id_field = "drupal_internal__tid"

    def normalize_source_object(self, term):
        return {
            "uuid": term.farmos_uuid or self.app.make_true_uuid(),
            "name": term.name,
            "description": term.description,
            "_src_object": term,
        }


class EquipmentTypeImporter(
    FromWuttaFarmTaxonomy, farmos_importing.model.EquipmentTypeImporter
):
    """
    WuttaFarm → farmOS API exporter for Equipment Types
    """

    source_model_class = model.EquipmentType


class EquipmentAssetImporter(
    FromWuttaFarmAsset, farmos_importing.model.EquipmentAssetImporter
):
    """
    WuttaFarm → farmOS API exporter for Equipment Assets
    """

    source_model_class = model.EquipmentAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())

        print(fields)
        fields.extend(
            [
                "manufacturer",
                "model",
                "serial_number",
                "equipment_type_uuids",
            ]
        )
        return fields

    def normalize_source_object(self, equipment):
        data = super().normalize_source_object(equipment)
        data.update(
            {
                "manufacturer": equipment.manufacturer,
                "model": equipment.model,
                "serial_number": equipment.serial_number,
                "equipment_type_uuids": [
                    etype.farmos_uuid for etype in equipment.equipment_types
                ],
            }
        )
        return data


class AnimalTypeImporter(
    FromWuttaFarmTaxonomy, farmos_importing.model.AnimalTypeImporter
):
    """
    WuttaFarm → farmOS API exporter for Animal Types
    """

    source_model_class = model.AnimalType


class MaterialTypeImporter(
    FromWuttaFarmTaxonomy, farmos_importing.model.MaterialTypeImporter
):
    """
    WuttaFarm → farmOS API exporter for Material Types
    """

    source_model_class = model.MaterialType


class UnitImporter(FromWuttaFarm, farmos_importing.model.UnitImporter):
    """
    WuttaFarm → farmOS API exporter for Units
    """

    source_model_class = model.Unit

    supported_fields = [
        "uuid",
        "name",
    ]

    drupal_internal_id_field = "drupal_internal__tid"

    def normalize_source_object(self, unit):
        return {
            "uuid": unit.farmos_uuid or self.app.make_true_uuid(),
            "name": unit.name,
            "_src_object": unit,
        }


class GroupAssetImporter(FromWuttaFarmAsset, farmos_importing.model.GroupAssetImporter):
    """
    WuttaFarm → farmOS API exporter for Group Assets
    """

    source_model_class = model.GroupAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "produces_eggs",
            ]
        )
        return fields

    def normalize_source_object(self, group):
        data = super().normalize_source_object(group)
        data.update(
            {
                "produces_eggs": group.produces_eggs,
            }
        )
        return data


class LandAssetImporter(FromWuttaFarmAsset, farmos_importing.model.LandAssetImporter):
    """
    WuttaFarm → farmOS API exporter for Land Assets
    """

    source_model_class = model.LandAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "land_type_id",
            ]
        )
        return fields

    def normalize_source_object(self, land):
        data = super().normalize_source_object(land)
        data.update(
            {
                "land_type_id": land.land_type.drupal_id,
            }
        )
        return data


class PlantTypeImporter(FromWuttaFarm, farmos_importing.model.PlantTypeImporter):
    """
    WuttaFarm → farmOS API exporter for Plant Types
    """

    source_model_class = model.PlantType

    supported_fields = [
        "uuid",
        "name",
    ]

    drupal_internal_id_field = "drupal_internal__tid"

    def normalize_source_object(self, plant_type):
        return {
            "uuid": plant_type.farmos_uuid or self.app.make_true_uuid(),
            "name": plant_type.name,
            "_src_object": plant_type,
        }


class SeasonImporter(FromWuttaFarm, farmos_importing.model.SeasonImporter):
    """
    WuttaFarm → farmOS API exporter for Seasons
    """

    source_model_class = model.Season

    supported_fields = [
        "uuid",
        "name",
    ]

    drupal_internal_id_field = "drupal_internal__tid"

    def normalize_source_object(self, season):
        return {
            "uuid": season.farmos_uuid or self.app.make_true_uuid(),
            "name": season.name,
            "_src_object": season,
        }


class PlantAssetImporter(FromWuttaFarmAsset, farmos_importing.model.PlantAssetImporter):
    """
    WuttaFarm → farmOS API exporter for Plant Assets
    """

    source_model_class = model.PlantAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "plant_type_uuids",
                "season_uuids",
            ]
        )
        return fields

    def normalize_source_object(self, plant):
        data = super().normalize_source_object(plant)
        data.update(
            {
                "plant_type_uuids": [pt.farmos_uuid for pt in plant.plant_types],
                "season_uuids": [s.farmos_uuid for s in plant.seasons],
            }
        )
        return data


class StructureAssetImporter(
    FromWuttaFarmAsset, farmos_importing.model.StructureAssetImporter
):
    """
    WuttaFarm → farmOS API exporter for Structure Assets
    """

    source_model_class = model.StructureAsset

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "structure_type_id",
            ]
        )
        return fields

    def normalize_source_object(self, structure):
        data = super().normalize_source_object(structure)
        data.update(
            {
                "structure_type_id": structure.structure_type.drupal_id,
            }
        )
        return data


class WaterAssetImporter(FromWuttaFarmAsset, farmos_importing.model.WaterAssetImporter):
    """
    WuttaFarm → farmOS API exporter for Water Assets
    """

    source_model_class = model.WaterAsset


##############################
# quantity importers
##############################


class FromWuttaFarmQuantity(FromWuttaFarm):
    """
    Base class for WuttaFarm -> farmOS quantity importers
    """

    supported_fields = [
        "uuid",
        "measure",
        "value_numerator",
        "value_denominator",
        "label",
        "quantity_type_uuid",
        "unit_uuid",
    ]

    def normalize_source_object(self, qty):
        return {
            "uuid": qty.farmos_uuid or self.app.make_true_uuid(),
            "measure": qty.measure_id,
            "value_numerator": qty.value_numerator,
            "value_denominator": qty.value_denominator,
            "label": qty.label,
            "quantity_type_uuid": qty.quantity_type.farmos_uuid,
            "unit_uuid": qty.units.farmos_uuid,
            "_src_object": qty,
        }


class MaterialQuantityImporter(
    FromWuttaFarmQuantity, farmos_importing.model.MaterialQuantityImporter
):
    """
    WuttaFarm → farmOS API exporter for Material Quantities
    """

    source_model_class = model.MaterialQuantity

    def normalize_source_object(self, quantity):
        data = super().normalize_source_object(quantity)

        if "material_types" in self.fields:
            data["material_types"] = [mt.farmos_uuid for mt in quantity.material_types]

        return data


class StandardQuantityImporter(
    FromWuttaFarmQuantity, farmos_importing.model.StandardQuantityImporter
):
    """
    WuttaFarm → farmOS API exporter for Standard Quantities
    """

    source_model_class = model.StandardQuantity


##############################
# log importers
##############################


class FromWuttaFarmLog(FromWuttaFarm):
    """
    Base class for WuttaFarm -> farmOS log importers
    """

    supported_fields = [
        "uuid",
        "name",
        "timestamp",
        "is_movement",
        "is_group_assignment",
        "status",
        "notes",
        "quick",
        "assets",
        "locations",
        "groups",
        "quantities",
    ]

    def normalize_source_object(self, log):
        return {
            "uuid": log.farmos_uuid or self.app.make_true_uuid(),
            "name": log.message,
            "timestamp": log.timestamp,
            "is_movement": log.is_movement,
            "is_group_assignment": log.is_group_assignment,
            "status": log.status,
            "notes": log.notes,
            "quick": self.config.parse_list(log.quick) if log.quick else [],
            "assets": [(a.asset_type, a.farmos_uuid) for a in log.assets],
            "locations": [(l.asset_type, l.farmos_uuid) for l in log.locations],
            "groups": [(g.asset_type, g.farmos_uuid) for g in log.groups],
            "quantities": [qty.farmos_uuid for qty in log.quantities],
            "_src_object": log,
        }


class ActivityLogImporter(FromWuttaFarmLog, farmos_importing.model.ActivityLogImporter):
    """
    WuttaFarm → farmOS API exporter for Activity Logs
    """

    source_model_class = model.ActivityLog


class HarvestLogImporter(FromWuttaFarmLog, farmos_importing.model.HarvestLogImporter):
    """
    WuttaFarm → farmOS API exporter for Harvest Logs
    """

    source_model_class = model.HarvestLog


class MedicalLogImporter(FromWuttaFarmLog, farmos_importing.model.MedicalLogImporter):
    """
    WuttaFarm → farmOS API exporter for Medical Logs
    """

    source_model_class = model.MedicalLog

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "vet",
            ]
        )
        return fields

    def normalize_source_object(self, log):
        data = super().normalize_source_object(log)
        data.update(
            {
                "vet": log.vet,
            }
        )
        return data


class ObservationLogImporter(
    FromWuttaFarmLog, farmos_importing.model.ObservationLogImporter
):
    """
    WuttaFarm → farmOS API exporter for Observation Logs
    """

    source_model_class = model.ObservationLog


class SeedingLogImporter(FromWuttaFarmLog, farmos_importing.model.SeedingLogImporter):
    """
    WuttaFarm → farmOS API exporter for Seeding Logs
    """

    source_model_class = model.SeedingLog

    def normalize_source_object(self, log):
        data = super().normalize_source_object(log)
        data.update(
            {
                "source": log.source,
                "purchase_date": log.purchase_date,
                "lot_number": log.lot_number,
            }
        )
        return data



"""add Standard Quantities

Revision ID: 5b6c87d8cddf
Revises: 1f98d27cabeb
Create Date: 2026-02-19 15:42:19.691148

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "5b6c87d8cddf"
down_revision: Union[str, None] = "1f98d27cabeb"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # measure
    op.create_table(
        "measure",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.Column("drupal_id", sa.String(length=20), nullable=True),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_measure")),
        sa.UniqueConstraint("drupal_id", name=op.f("uq_measure_drupal_id")),
        sa.UniqueConstraint("name", name=op.f("uq_measure_name")),
    )
    op.create_table(
        "measure_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column("name", sa.String(length=100), autoincrement=False, nullable=True),
        sa.Column(
            "drupal_id", sa.String(length=20), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_measure_version")
        ),
    )
    op.create_index(
        op.f("ix_measure_version_end_transaction_id"),
        "measure_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_measure_version_operation_type"),
        "measure_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_measure_version_pk_transaction_id",
        "measure_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_measure_version_pk_validity",
        "measure_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_measure_version_transaction_id"),
        "measure_version",
        ["transaction_id"],
        unique=False,
    )

    # quantity
    op.create_table(
        "quantity",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("quantity_type_id", sa.String(length=50), nullable=False),
        sa.Column("measure_id", sa.String(length=20), nullable=False),
        sa.Column("value_numerator", sa.Integer(), nullable=False),
        sa.Column("value_denominator", sa.Integer(), nullable=False),
        sa.Column("units_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("label", sa.String(length=255), nullable=True),
        sa.Column("farmos_uuid", wuttjamaican.db.util.UUID(), nullable=True),
        sa.Column("drupal_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["measure_id"],
            ["measure.drupal_id"],
            name=op.f("fk_quantity_measure_id_measure"),
        ),
        sa.ForeignKeyConstraint(
            ["quantity_type_id"],
            ["quantity_type.drupal_id"],
            name=op.f("fk_quantity_quantity_type_id_quantity_type"),
        ),
        sa.ForeignKeyConstraint(
            ["units_uuid"], ["unit.uuid"], name=op.f("fk_quantity_units_uuid_unit")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_quantity")),
        sa.UniqueConstraint("drupal_id", name=op.f("uq_quantity_drupal_id")),
        sa.UniqueConstraint("farmos_uuid", name=op.f("uq_quantity_farmos_uuid")),
    )
    op.create_table(
        "quantity_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "quantity_type_id", sa.String(length=50), autoincrement=False, nullable=True
        ),
        sa.Column(
            "measure_id", sa.String(length=20), autoincrement=False, nullable=True
        ),
        sa.Column("value_numerator", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column(
            "value_denominator", sa.Integer(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "units_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("label", sa.String(length=255), autoincrement=False, nullable=True),
        sa.Column(
            "farmos_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("drupal_id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_quantity_version")
        ),
    )
    op.create_index(
        op.f("ix_quantity_version_end_transaction_id"),
        "quantity_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_quantity_version_operation_type"),
        "quantity_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_quantity_version_pk_transaction_id",
        "quantity_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_quantity_version_pk_validity",
        "quantity_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_quantity_version_transaction_id"),
        "quantity_version",
        ["transaction_id"],
        unique=False,
    )

    # quantity_standard
    op.create_table(
        "quantity_standard",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["uuid"], ["quantity.uuid"], name=op.f("fk_quantity_standard_uuid_quantity")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_quantity_standard")),
    )
    op.create_table(
        "quantity_standard_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_quantity_standard_version")
        ),
    )
    op.create_index(
        op.f("ix_quantity_standard_version_end_transaction_id"),
        "quantity_standard_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_quantity_standard_version_operation_type"),
        "quantity_standard_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_quantity_standard_version_pk_transaction_id",
        "quantity_standard_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_quantity_standard_version_pk_validity",
        "quantity_standard_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_quantity_standard_version_transaction_id"),
        "quantity_standard_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # quantity_standard
    op.drop_index(
        op.f("ix_quantity_standard_version_transaction_id"),
        table_name="quantity_standard_version",
    )
    op.drop_index(
        "ix_quantity_standard_version_pk_validity",
        table_name="quantity_standard_version",
    )
    op.drop_index(
        "ix_quantity_standard_version_pk_transaction_id",
        table_name="quantity_standard_version",
    )
    op.drop_index(
        op.f("ix_quantity_standard_version_operation_type"),
        table_name="quantity_standard_version",
    )
    op.drop_index(
        op.f("ix_quantity_standard_version_end_transaction_id"),
        table_name="quantity_standard_version",
    )
    op.drop_table("quantity_standard_version")
    op.drop_table("quantity_standard")

    # quantity
    op.drop_index(
        op.f("ix_quantity_version_transaction_id"), table_name="quantity_version"
    )
    op.drop_index("ix_quantity_version_pk_validity", table_name="quantity_version")
    op.drop_index(
        "ix_quantity_version_pk_transaction_id", table_name="quantity_version"
    )
    op.drop_index(
        op.f("ix_quantity_version_operation_type"), table_name="quantity_version"
    )
    op.drop_index(
        op.f("ix_quantity_version_end_transaction_id"), table_name="quantity_version"
    )
    op.drop_table("quantity_version")
    op.drop_table("quantity")

    # measure
    op.drop_index(
        op.f("ix_measure_version_transaction_id"), table_name="measure_version"
    )
    op.drop_index("ix_measure_version_pk_validity", table_name="measure_version")
    op.drop_index("ix_measure_version_pk_transaction_id", table_name="measure_version")
    op.drop_index(
        op.f("ix_measure_version_operation_type"), table_name="measure_version"
    )
    op.drop_index(
        op.f("ix_measure_version_end_transaction_id"), table_name="measure_version"
    )
    op.drop_table("measure_version")
    op.drop_table("measure")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model definition for Animal Types
"""

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.db import model

from wuttafarm.db.model.asset import AssetMixin, add_asset_proxies, EggMixin


class AnimalType(model.Base):
    """
    Represents an "animal type" (taxonomy term) from farmOS
    """

    __tablename__ = "animal_type"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Animal Type",
        "model_title_plural": "Animal Types",
    }

    uuid = model.uuid_column()

    name = sa.Column(
        sa.String(length=100),
        nullable=False,
        doc="""
        Name of the animal type.
        """,
    )

    description = sa.Column(
        sa.String(length=255),
        nullable=True,
        doc="""
        Optional description for the animal type.
        """,
    )

    farmos_uuid = sa.Column(
        model.UUID(),
        nullable=True,
        unique=True,
        doc="""
        UUID for the animal type within farmOS.
        """,
    )

    drupal_id = sa.Column(
        sa.Integer(),
        nullable=True,
        unique=True,
        doc="""
        Drupal internal ID for the animal type.
        """,
    )

    animal_assets = orm.relationship(
        "AnimalAsset",
        doc="""
        List of animal assets of this type.
        """,
        back_populates="animal_type",
    )

    def __str__(self):
        return self.name or ""


class AnimalAsset(AssetMixin, EggMixin, model.Base):
    """
    Represents an animal asset from farmOS
    """

    __tablename__ = "asset_animal"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Animal Asset",
        "model_title_plural": "Animal Assets",
        "farmos_asset_type": "animal",
    }

    animal_type_uuid = model.uuid_fk_column("animal_type.uuid", nullable=False)
    animal_type = orm.relationship(
        "AnimalType",
        doc="""
        Reference to the animal type.
        """,
        back_populates="animal_assets",
    )

    birthdate = sa.Column(
        sa.DateTime(),
        nullable=True,
        doc="""
        Birth date (and time) for the animal, if known.
        """,
    )

    sex = sa.Column(
        sa.String(length=1),
        nullable=True,
        doc="""
        Sex of the animal.
        """,
    )

    is_sterile = sa.Column(
        sa.Boolean(),
        nullable=True,
        doc="""
        Whether the animal is sterile (e.g. castrated).
        """,
    )


add_asset_proxies(AnimalAsset)



"""add produces_eggs via EggMixin

Revision ID: 82a03f4ef1a4
Revises: 11e0e46f48a6
Create Date: 2026-02-18 18:45:36.015144

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "82a03f4ef1a4"
down_revision: Union[str, None] = "11e0e46f48a6"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # asset_animal
    op.add_column(
        "asset_animal", sa.Column("produces_eggs", sa.Boolean(), nullable=True)
    )
    op.add_column(
        "asset_animal_version",
        sa.Column("produces_eggs", sa.Boolean(), autoincrement=False, nullable=True),
    )

    # asset_group
    op.add_column(
        "asset_group", sa.Column("produces_eggs", sa.Boolean(), nullable=True)
    )
    op.add_column(
        "asset_group_version",
        sa.Column("produces_eggs", sa.Boolean(), autoincrement=False, nullable=True),
    )


def downgrade() -> None:

    # asset_group
    op.drop_column("asset_group_version", "produces_eggs")
    op.drop_column("asset_group", "produces_eggs")

    # asset_animal
    op.drop_column("asset_animal_version", "produces_eggs")
    op.drop_column("asset_animal", "produces_eggs")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for farmOS Equipment
"""

from webhelpers2.html import tags

from wuttafarm.web.views.farmos.master import TaxonomyMasterView
from wuttafarm.web.views.farmos.assets import AssetMasterView
from wuttafarm.web.forms.schema import FarmOSEquipmentTypeRefs


class EquipmentTypeView(TaxonomyMasterView):
    """
    Master view for Equipment Types in farmOS.
    """

    model_name = "farmos_equipment_type"
    model_title = "farmOS Equipment Type"
    model_title_plural = "farmOS Equipment Types"

    route_prefix = "farmos_equipment_types"
    url_prefix = "/farmOS/equipment-types"

    farmos_taxonomy_type = "equipment_type"
    farmos_refurl_path = "/admin/structure/taxonomy/manage/equipment_type/overview"

    def get_xref_buttons(self, equipment_type):
        buttons = super().get_xref_buttons(equipment_type)
        model = self.app.model
        session = self.Session()

        if wf_equipment_type := (
            session.query(model.EquipmentType)
            .filter(model.EquipmentType.farmos_uuid == equipment_type["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "equipment_types.view", uuid=wf_equipment_type.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


class EquipmentAssetView(AssetMasterView):
    """
    Master view for farmOS Equipment Assets
    """

    model_name = "farmos_equipment_assets"
    model_title = "farmOS Equipment Asset"
    model_title_plural = "farmOS Equipment Assets"

    route_prefix = "farmos_equipment_assets"
    url_prefix = "/farmOS/assets/equipment"

    farmos_asset_type = "equipment"
    farmos_refurl_path = "/assets/equipment"

    labels = {
        "equipment_types": "Equipment Type",
    }

    grid_columns = [
        "thumbnail",
        "drupal_id",
        "name",
        "equipment_types",
        "manufacturer",
        "model",
        "serial_number",
        "groups",
        "owners",
        "archived",
    ]

    def get_farmos_api_includes(self):
        includes = super().get_farmos_api_includes()
        includes.update(["equipment_type"])
        return includes

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # equipment_types
        g.set_renderer("equipment_types", self.render_equipment_types_for_grid)

    def render_equipment_types_for_grid(self, equipment, field, value):
        if self.farmos_style_grid_links:
            links = []
            for etype in value:
                url = self.request.route_url(
                    f"farmos_equipment_types.view", uuid=etype["uuid"]
                )
                links.append(tags.link_to(etype["name"], url))
            return ", ".join(links)

        return ", ".join([etype["name"] for etype in value])

    def normalize_asset(self, equipment, included):
        data = super().normalize_asset(equipment, included)

        equipment_type_objects = []
        rels = equipment["relationships"]
        for etype in rels["equipment_type"]["data"]:
            uuid = etype["id"]
            equipment_type = {
                "uuid": uuid,
                "type": etype["type"],
            }
            if etype := included.get(uuid):
                equipment_type.update(
                    {
                        "name": etype["attributes"]["name"],
                    }
                )
            equipment_type_objects.append(equipment_type)

        data.update(
            {
                "manufacturer": equipment["attributes"]["manufacturer"],
                "model": equipment["attributes"]["model"],
                "serial_number": equipment["attributes"]["serial_number"],
                "equipment_types": equipment_type_objects,
            }
        )

        return data

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # equipment_types
        f.fields.insert_after("name", "equipment_types")
        f.set_node("equipment_types", FarmOSEquipmentTypeRefs(self.request))

        # manufacturer
        f.fields.insert_after("equipment_types", "manufacturer")

        # model
        f.fields.insert_after("manufacturer", "model")

        # serial_number
        f.fields.insert_after("model", "serial_number")

    def get_xref_buttons(self, equipment):
        model = self.app.model
        session = self.Session()

        buttons = super().get_xref_buttons(equipment)

        if wf_equipment := (
            session.query(model.Asset)
            .filter(model.Asset.farmos_uuid == equipment["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "equipment_assets.view", uuid=wf_equipment.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    EquipmentTypeView = kwargs.get("EquipmentTypeView", base["EquipmentTypeView"])
    EquipmentTypeView.defaults(config)

    EquipmentAssetView = kwargs.get("EquipmentAssetView", base["EquipmentAssetView"])
    EquipmentAssetView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
"""
Tasks for WuttaFarm
"""

import os
import shutil

from invoke import task


@task
def release(c, skip_tests=False):
    """
    Release a new version of WuttaFarm
    """
    if not skip_tests:
        # TODO
        # c.run("pytest")
        pass

    if os.path.exists("dist"):
        shutil.rmtree("dist")

    c.run("python -m build")
    c.run("twine upload dist/*")


@task
def update_style(c):
    """
    Build/update the `wuttafarm-buefy.css` file
    """
    os.chdir("style")
    # c.run("nvm use lts/krypton")
    c.run("npm install")
    c.run("npm run build")
    os.chdir(os.pardir)
    shutil.copy("style/dist/css/wuttafarm-buefy.css", "src/wuttafarm/web/static/css/")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
View for farmOS Structures
"""

import datetime

import colander

from wuttaweb.forms.schema import WuttaDateTime
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views.farmos import FarmOSMasterView
from wuttafarm.web.forms.widgets import ImageWidget


class StructureView(FarmOSMasterView):
    """
    View for farmOS Structures
    """

    model_name = "farmos_structure_asset"
    model_title = "farmOS Structure"
    model_title_plural = "farmOS Structures"

    route_prefix = "farmos_structure_assets"
    url_prefix = "/farmOS/structures"

    farmos_refurl_path = "/assets/structure"

    grid_columns = [
        "name",
        "archived",
        "created",
        "changed",
    ]

    sort_defaults = "name"

    form_fields = [
        "name",
        "archived",
        "structure_type",
        "is_location",
        "is_fixed",
        "notes",
        "created",
        "changed",
        "raw_image_url",
        "large_image_url",
        "thumbnail_image_url",
        "image",
    ]

    def get_grid_data(self, columns=None, session=None):
        structures = self.farmos_client.resource.get("asset", "structure")
        return [self.normalize_structure(s) for s in structures["data"]]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")
        g.set_searchable("name")

        # created
        g.set_renderer("created", "datetime")

        # changed
        g.set_renderer("changed", "datetime")

        # archived
        g.set_renderer("archived", "boolean")

    def get_instance(self):
        structure = self.farmos_client.resource.get_id(
            "asset", "structure", self.request.matchdict["uuid"]
        )
        self.raw_json = structure
        data = self.normalize_structure(structure["data"])

        if relationships := structure["data"].get("relationships"):

            # add owners
            if owner := relationships.get("owner"):
                data["owners"] = []
                for owner_data in owner["data"]:
                    owner = self.farmos_client.resource.get_id(
                        "user", "user", owner_data["id"]
                    )
                    data["owners"].append(
                        {
                            "uuid": owner["data"]["id"],
                            "display_name": owner["data"]["attributes"]["display_name"],
                        }
                    )

            # add image urls
            if image := relationships.get("image"):
                if image["data"]:
                    image = self.farmos_client.resource.get_id(
                        "file", "file", image["data"][0]["id"]
                    )
                    data["raw_image_url"] = self.app.get_farmos_url(
                        image["data"]["attributes"]["uri"]["url"]
                    )
                    # nb. other styles available:  medium, wide
                    data["large_image_url"] = image["data"]["attributes"][
                        "image_style_uri"
                    ]["large"]
                    data["thumbnail_image_url"] = image["data"]["attributes"][
                        "image_style_uri"
                    ]["thumbnail"]

        return data

    def get_instance_title(self, structure):
        return structure["name"]

    def normalize_structure(self, structure):

        if created := structure["attributes"].get("created"):
            created = datetime.datetime.fromisoformat(created)
            created = self.app.localtime(created)

        if changed := structure["attributes"].get("changed"):
            changed = datetime.datetime.fromisoformat(changed)
            changed = self.app.localtime(changed)

        if self.farmos_4x:
            archived = structure["attributes"]["archived"]
        else:
            archived = structure["attributes"]["status"] == "archived"

        return {
            "uuid": structure["id"],
            "drupal_id": structure["attributes"]["drupal_internal__id"],
            "name": structure["attributes"]["name"],
            "structure_type": structure["attributes"]["structure_type"],
            "is_fixed": structure["attributes"]["is_fixed"],
            "is_location": structure["attributes"]["is_location"],
            "notes": structure["attributes"]["notes"] or colander.null,
            "archived": archived,
            "created": created,
            "changed": changed,
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        structure = f.model_instance

        # is_fixed
        f.set_node("is_fixed", colander.Boolean())

        # is_location
        f.set_node("is_location", colander.Boolean())

        # notes
        f.set_widget("notes", "notes")

        # created
        f.set_node("created", WuttaDateTime())
        f.set_widget("created", WuttaDateTimeWidget(self.request))

        # changed
        f.set_node("changed", WuttaDateTime())
        f.set_widget("changed", WuttaDateTimeWidget(self.request))

        # archived
        f.set_node("archived", colander.Boolean())

        # image
        if url := structure.get("large_image_url"):
            f.set_widget("image", ImageWidget("structure image"))
            f.set_default("image", url)

    def get_xref_buttons(self, structure):
        model = self.app.model
        session = self.Session()

        buttons = [
            self.make_button(
                "View in farmOS",
                primary=True,
                url=self.app.get_farmos_url(f"/asset/{structure['drupal_id']}"),
                target="_blank",
                icon_left="external-link-alt",
            ),
        ]

        if wf_structure := (
            session.query(model.StructureAsset)
            .filter(model.StructureAsset.farmos_uuid == structure["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "structure_assets.view", uuid=wf_structure.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    StructureView = kwargs.get("StructureView", base["StructureView"])
    StructureView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaFarm config extensions
"""

import os

from wuttjamaican.conf import WuttaConfigExtension


class WuttaFarmConfig(WuttaConfigExtension):
    """
    Config extension for WuttaFarm
    """

    key = "wuttafarm"

    def configure(self, config):

        # app info
        config.setdefault(f"{config.appname}.app_title", "WuttaFarm")
        config.setdefault(f"{config.appname}.app_dist", "WuttaFarm")

        # app model/enum
        config.setdefault(f"{config.appname}.model_spec", "wuttafarm.db.model")
        config.setdefault(f"{config.appname}.enum_spec", "wuttafarm.enum")

        # app handler
        config.setdefault(
            f"{config.appname}.app.handler", "wuttafarm.app:WuttaFarmAppHandler"
        )

        # web app stuff
        config.setdefault(
            f"{config.appname}.web.menus.handler.default_spec",
            "wuttafarm.web.menus:WuttaFarmMenuHandler",
        )
        config.setdefault("wuttaweb.grids.default_pagesize", "50")

        # web app libcache
        # config.setdefault('wuttaweb.static_libcache.module', 'wuttafarm.web.static')

        # maybe override cert validation for requests lib.
        # nb. this is "global" and not "specific" to the farmos API requests!
        if bundle := config.get(f"{config.appname}.requests_ca_bundle"):
            os.environ.setdefault("REQUESTS_CA_BUNDLE", bundle)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Custom grid stuff for use with farmOS / JSONAPI
"""

import datetime

from wuttaweb.grids.filters import GridFilter


class SimpleFilter(GridFilter):

    default_verbs = ["equal", "not_equal"]

    def __init__(self, request, key, path=None, **kwargs):
        super().__init__(request, key, **kwargs)
        self.path = path or key

    def filter_equal(self, data, value):
        if value := self.coerce_value(value):
            data.add_filter(self.path, "=", value)
        return data

    def filter_not_equal(self, data, value):
        if value := self.coerce_value(value):
            data.add_filter(self.path, "<>", value)
        return data

    def filter_is_null(self, data, value):
        data.add_filter(self.path, "IS NULL", None)
        return data

    def filter_is_not_null(self, data, value):
        data.add_filter(self.path, "IS NOT NULL", None)
        return data


class StringFilter(SimpleFilter):

    default_verbs = ["contains", "equal", "not_equal"]

    def filter_contains(self, data, value):
        if value := self.coerce_value(value):
            data.add_filter(self.path, "CONTAINS", value)
        return data


class NullableStringFilter(StringFilter):

    default_verbs = ["contains", "equal", "not_equal", "is_null", "is_not_null"]


class IntegerFilter(SimpleFilter):

    default_verbs = [
        "equal",
        "not_equal",
        "less_than",
        "less_equal",
        "greater_than",
        "greater_equal",
    ]

    def filter_less_than(self, data, value):
        if value := self.coerce_value(value):
            data.add_filter(self.path, "<", value)
        return data

    def filter_less_equal(self, data, value):
        if value := self.coerce_value(value):
            data.add_filter(self.path, "<=", value)
        return data

    def filter_greater_than(self, data, value):
        if value := self.coerce_value(value):
            data.add_filter(self.path, ">", value)
        return data

    def filter_greater_equal(self, data, value):
        if value := self.coerce_value(value):
            data.add_filter(self.path, ">=", value)
        return data


class NullableIntegerFilter(IntegerFilter):

    default_verbs = ["equal", "not_equal", "is_null", "is_not_null"]


class BooleanFilter(SimpleFilter):

    default_verbs = ["is_true", "is_false"]

    def filter_is_true(self, data, value):
        data.add_filter(self.path, "=", 1)
        return data

    def filter_is_false(self, data, value):
        data.add_filter(self.path, "=", 0)
        return data


class NullableBooleanFilter(BooleanFilter):

    default_verbs = ["is_true", "is_false", "is_null", "is_not_null"]


# TODO: this may not work, it's not used anywhere yet
class DateFilter(SimpleFilter):

    data_type = "date"

    default_verbs = [
        "equal",
        "not_equal",
        "greater_than",
        "greater_equal",
        "less_than",
        "less_equal",
        # 'between',
    ]

    default_verb_labels = {
        "equal": "on",
        "not_equal": "not on",
        "greater_than": "after",
        "greater_equal": "on or after",
        "less_than": "before",
        "less_equal": "on or before",
        # "between": "between",
        "is_null": "is null",
        "is_not_null": "is not null",
        "is_any": "is any",
    }

    def coerce_value(self, value):
        if value:
            if isinstance(value, datetime.date):
                return value

            try:
                dt = datetime.datetime.strptime(value, "%Y-%m-%d")
            except ValueError:
                log.warning("invalid date value: %s", value)
            else:
                return dt.date()

        return None


# TODO: this is not very complete yet, so far used only for animal birthdate
class DateTimeFilter(DateFilter):

    default_verbs = ["equal", "is_null", "is_not_null"]

    def coerce_value(self, value):
        """
        Convert user input to a proper ``datetime.date`` object.
        """
        if value:
            if isinstance(value, datetime.date):
                return value

            try:
                dt = datetime.datetime.strptime(value, "%Y-%m-%d")
            except ValueError:
                log.warning("invalid date value: %s", value)
            else:
                return dt.date()

        return None

    def filter_equal(self, data, value):
        if value := self.coerce_value(value):

            start = datetime.datetime.combine(value, datetime.time(0))
            start = self.app.localtime(start, from_utc=False)

            stop = datetime.datetime.combine(
                value + datetime.timedelta(days=1), datetime.time(0)
            )
            stop = self.app.localtime(stop, from_utc=False)

            data.add_filter(self.path, ">=", int(start.timestamp()))
            data.add_filter(self.path, "<", int(stop.timestamp()))

        return data


class SimpleSorter:

    def __init__(self, key):
        self.key = key

    def __call__(self, data, sortdir):
        data.add_sorter(self.key, sortdir)
        return data


class ResourceData:

    def __init__(
        self,
        config,
        farmos_client,
        content_type,
        include=None,
        normalizer=None,
    ):
        self.config = config
        self.farmos_client = farmos_client
        self.entity, self.bundle = content_type.split("--")
        self.filters = []
        self.sorters = []
        self.include = include
        self.normalizer = normalizer
        self._data = None

    def __bool__(self):
        return True

    def __getitem__(self, subscript):
        return self.get_data()[subscript]

    def __len__(self):
        return len(self._data)

    def add_filter(self, path, operator, value):
        self.filters.append((path, operator, value))

    def add_sorter(self, path, sortdir):
        self.sorters.append((path, sortdir))

    def get_data(self):
        if self._data is None:
            params = {}

            i = 0
            for path, operator, value in self.filters:
                i += 1
                key = f"{i:03d}"
                params[f"filter[{key}][condition][path]"] = path
                params[f"filter[{key}][condition][operator]"] = operator
                params[f"filter[{key}][condition][value]"] = value

            sorters = []
            for path, sortdir in self.sorters:
                prefix = "-" if sortdir == "desc" else ""
                sorters.append(f"{prefix}{path}")
            if sorters:
                params["sort"] = ",".join(sorters)

            # nb. while the API allows for pagination, it does not
            # tell me how many total records there are (IIUC).  also
            # if i ask for e.g. items 21-40 (page 2 @ 20/page) i am
            # not guaranteed to get 20 items even if there are plenty
            # in the DB, since Drupal may filter some out based on
            # permissions.  (granted that may not be an issue in
            # practice, but can't rule it out.) so the punchline is,
            # we fetch "all" (sic) data and send it to the frontend,
            # and pagination happens there.

            # TODO: if we ever try again, this sort of works...
            # params["page[offset]"] = start
            # params["page[limit]"] = stop - start

            if self.include:
                params["include"] = self.include

            result = self.farmos_client.resource.get(
                self.entity, self.bundle, params=params
            )
            data = result["data"]
            included = {obj["id"]: obj for obj in result.get("included", [])}

            if self.normalizer:
                data = [self.normalizer(d, included) for d in data]

            self._data = data
        return self._data



"""remove unwanted unique constraint

Revision ID: 5f474125a80e
Revises: 0771322957bd
Create Date: 2026-03-04 12:03:16.034291

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "5f474125a80e"
down_revision: Union[str, None] = "0771322957bd"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # asset_land
    op.drop_constraint(
        op.f("uq_asset_land_land_type_uuid"), "asset_land", type_="unique"
    )


def downgrade() -> None:

    # asset_land
    op.create_unique_constraint(
        op.f("uq_asset_land_land_type_uuid"),
        "asset_land",
        ["land_type_uuid"],
        postgresql_nulls_not_distinct=False,
    )



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Quantities
"""

from collections import OrderedDict

from webhelpers2.html import tags

from wuttaweb.db import Session

from wuttafarm.web.views import WuttaFarmMasterView
from wuttafarm.db.model import (
    QuantityType,
    Quantity,
    StandardQuantity,
    MaterialQuantity,
)
from wuttafarm.web.forms.schema import UnitRef, LogRef, MaterialTypeRefs
from wuttafarm.util import get_log_type_enum


def get_quantity_type_enum(config):
    app = config.get_app()
    model = app.model
    session = Session()
    quantity_types = OrderedDict()
    query = session.query(model.QuantityType).order_by(model.QuantityType.name)
    for quantity_type in query:
        quantity_types[quantity_type.drupal_id] = quantity_type.name
    return quantity_types


class QuantityTypeView(WuttaFarmMasterView):
    """
    Master view for Quantity Types
    """

    model_class = QuantityType
    route_prefix = "quantity_types"
    url_prefix = "/quantity-types"

    grid_columns = [
        "name",
        "description",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "drupal_id",
        "farmos_uuid",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_xref_buttons(self, quantity_type):
        buttons = super().get_xref_buttons(quantity_type)

        if quantity_type.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_quantity_types.view", uuid=quantity_type.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


class QuantityMasterView(WuttaFarmMasterView):
    """
    Base class for Quantity master views
    """

    farmos_entity_type = "quantity"

    labels = {
        "log_id": "Log ID",
    }

    grid_columns = [
        "drupal_id",
        "log_id",
        "log_status",
        "log_timestamp",
        "log_type",
        "log_name",
        "log_assets",
        "measure",
        "value",
        "units",
        "label",
        "quantity_type",
    ]

    sort_defaults = ("log_timestamp", "desc")

    form_fields = [
        "quantity_type",
        "as_text",
        "measure",
        "value",
        "units",
        "label",
        "log",
        "drupal_id",
        "farmos_uuid",
    ]

    def get_query(self, session=None):
        """ """
        model = self.app.model
        model_class = self.get_model_class()
        session = session or self.Session()

        query = session.query(model_class)
        if model_class is not model.Quantity:
            query = query.join(model.Quantity)

        query = query.join(model.Measure).join(model.Unit)

        query = query.outerjoin(model.LogQuantity).outerjoin(model.Log)

        return query

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        model = self.app.model
        model_class = self.get_model_class()
        session = self.Session()

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)
        g.set_sorter("drupal_id", model.Quantity.drupal_id)

        # log_id
        g.set_renderer("log_id", self.render_log_id_for_grid)
        g.set_sorter("log_id", model.Log.drupal_id)

        # log_status
        g.set_renderer("log_status", self.render_log_status_for_grid)
        g.set_sorter("log_status", model.Log.status)

        # log_timestamp
        g.set_renderer("log_timestamp", self.render_log_timestamp_for_grid)
        g.set_sorter("log_timestamp", model.Log.timestamp)

        # log_type
        self.log_type_enum = get_log_type_enum(self.config, session)
        g.set_renderer("log_type", self.render_log_type_for_grid)
        g.set_sorter("log_type", model.Log.log_type)

        # log_name
        g.set_renderer("log_name", self.render_log_name_for_grid)
        g.set_sorter("log_name", model.Log.message)
        if not self.farmos_style_grid_links:
            g.set_link("log_name")

        # log_assets
        g.set_renderer("log_assets", self.render_log_assets_for_grid)
        if not self.farmos_style_grid_links:
            g.set_link("log_assets")

        # quantity_type
        if model_class is not model.Quantity:
            g.remove("quantity_type")
        else:
            g.set_enum("quantity_type", get_quantity_type_enum(self.config))

        # measure
        g.set_sorter("measure", model.Measure.name)

        # value
        g.set_renderer("value", self.render_value_for_grid)

        # units
        g.set_sorter("units", model.Unit.name)

        # label
        g.set_sorter("label", model.Quantity.label)

        # view action links to final quantity record
        if model_class is model.Quantity:

            def quantity_url(quantity, i):
                return self.request.route_url(
                    f"quantities_{quantity.quantity_type_id}.view", uuid=quantity.uuid
                )

            g.add_action("view", icon="eye", url=quantity_url)

    def render_log_id_for_grid(self, quantity, field, value):
        if log := quantity.log:
            return log.drupal_id
        return None

    def render_log_status_for_grid(self, quantity, field, value):
        enum = self.app.enum
        if log := quantity.log:
            return enum.LOG_STATUS.get(log.status, log.status)
        return None

    def render_log_timestamp_for_grid(self, quantity, field, value):
        if log := quantity.log:
            return self.app.render_date(log.timestamp)
        return None

    def render_log_type_for_grid(self, quantity, field, value):
        if log := quantity.log:
            return self.log_type_enum.get(log.log_type, log.log_type)
        return None

    def render_log_name_for_grid(self, quantity, field, value):
        if log := quantity.log:
            if self.farmos_style_grid_links:
                url = self.request.route_url(f"logs_{log.log_type}.view", uuid=log.uuid)
                return tags.link_to(log.message, url)
            return log.message
        return None

    def render_log_assets_for_grid(self, quantity, field, value):
        if log := quantity.log:
            if self.farmos_style_grid_links:
                links = []
                for asset in log.assets:
                    url = self.request.route_url(
                        f"{asset.asset_type}_assets.view", uuid=asset.uuid
                    )
                    links.append(tags.link_to(str(asset), url))
                return ", ".join(links)
            return [str(a) for a in log.assets]
        return None

    def render_value_for_grid(self, quantity, field, value):
        value = quantity.value_numerator / quantity.value_denominator
        return self.app.render_quantity(value)

    def get_instance_title(self, quantity):
        return quantity.render_as_text(self.config)

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        quantity = form.model_instance

        # as_text
        if self.creating or self.editing:
            f.remove("as_text")
        else:
            f.set_default("as_text", quantity.render_as_text(self.config))

        # quantity_type
        if self.creating:
            f.remove("quantity_type")
        else:
            f.set_readonly("quantity_type")
            f.set_default("quantity_type", quantity.quantity_type.name)

        # measure
        if self.creating:
            f.remove("measure")
        else:
            f.set_readonly("measure")
            f.set_default("measure", quantity.measure.name)

        # value
        if self.creating:
            f.remove("value")
        else:
            value = quantity.value_numerator / quantity.value_denominator
            value = self.app.render_quantity(value)
            f.set_default(
                "value",
                f"{value} ({quantity.value_numerator} / {quantity.value_denominator})",
            )

        # units
        if self.creating:
            f.remove("units")
        else:
            f.set_readonly("units")
            f.set_node("units", UnitRef(self.request))
            # TODO: ugh
            f.set_default("units", quantity.quantity.units)

        # log
        if self.creating or self.editing:
            f.remove("log")
        else:
            f.set_node("log", LogRef(self.request))
            f.set_default("log", quantity.log)

    def get_xref_buttons(self, quantity):
        buttons = super().get_xref_buttons(quantity)

        if quantity.farmos_uuid:
            url = self.request.route_url(
                f"farmos_quantities_{quantity.quantity_type_id}.view",
                uuid=quantity.farmos_uuid,
            )
            buttons.append(
                self.make_button(
                    "View farmOS record", primary=True, url=url, icon_left="eye"
                )
            )

        return buttons


class AllQuantityView(QuantityMasterView):
    """
    Master view for All Quantities
    """

    model_class = Quantity
    route_prefix = "quantities"
    url_prefix = "/quantities"

    viewable = False
    creatable = False
    editable = False
    deletable = False
    model_is_versioned = False

    farmos_refurl_path = "/log-quantities"


class StandardQuantityView(QuantityMasterView):
    """
    Master view for Standard Quantities
    """

    model_class = StandardQuantity
    route_prefix = "quantities_standard"
    url_prefix = "/quantities/standard"

    farmos_bundle = "standard"
    farmos_refurl_path = "/log-quantities/standard"


class MaterialQuantityView(QuantityMasterView):
    """
    Master view for Material Quantities
    """

    model_class = MaterialQuantity
    route_prefix = "quantities_material"
    url_prefix = "/quantities/material"

    farmos_bundle = "material"
    farmos_refurl_path = "/log-quantities/material"

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # material_types
        g.columns.append("material_types")
        g.set_label("material_types", "Material Type", column_only=True)
        g.set_renderer("material_types", self.render_material_types_for_grid)

    def render_material_types_for_grid(self, quantity, field, value):
        if self.farmos_style_grid_links:
            links = []
            for mtype in quantity.material_types:
                url = self.request.route_url("material_types.view", uuid=mtype.uuid)
                links.append(tags.link_to(str(mtype), url))
            return ", ".join(links)

        return ", ".join([str(mtype) for mtype in quantity.material_types])

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        quantity = form.model_instance

        # material_types
        f.fields.insert_after("quantity_type", "material_types")
        f.set_node("material_types", MaterialTypeRefs(self.request))
        if not self.creating:
            f.set_default("material_types", quantity.material_types)

    def objectify(self, form):
        quantity = super().objectify(form)
        data = form.validated

        self.set_material_types(quantity, data["material_types"])

        return quantity

    def set_material_types(self, quantity, desired):
        model = self.app.model
        session = self.Session()

        current = {mt.uuid.hex: mt for mt in quantity.material_types}

        for mtype in desired:
            if mtype["uuid"] not in current:
                mtype = session.get(model.MaterialType, mtype["uuid"])
                assert mtype
                quantity.material_types.append(mtype)

        desired = [mtype["uuid"] for mtype in desired]
        for uuid, mtype in current.items():
            if uuid not in desired:
                quantity.material_types.remove(mtype)


def defaults(config, **kwargs):
    base = globals()

    QuantityTypeView = kwargs.get("QuantityTypeView", base["QuantityTypeView"])
    QuantityTypeView.defaults(config)

    AllQuantityView = kwargs.get("AllQuantityView", base["AllQuantityView"])
    AllQuantityView.defaults(config)

    StandardQuantityView = kwargs.get(
        "StandardQuantityView", base["StandardQuantityView"]
    )
    StandardQuantityView.defaults(config)

    MaterialQuantityView = kwargs.get(
        "MaterialQuantityView", base["MaterialQuantityView"]
    )
    MaterialQuantityView.defaults(config)


def includeme(config):
    defaults(config)



"""add LogGroup

Revision ID: 74d32b4ec210
Revises: 3bef7d380a38
Create Date: 2026-02-28 21:35:24.125784

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "74d32b4ec210"
down_revision: Union[str, None] = "3bef7d380a38"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # log_group
    op.create_table(
        "log_group",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("log_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("asset_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["asset_uuid"], ["asset.uuid"], name=op.f("fk_log_group_asset_uuid_asset")
        ),
        sa.ForeignKeyConstraint(
            ["log_uuid"], ["log.uuid"], name=op.f("fk_log_group_log_uuid_log")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_log_group")),
    )
    op.create_table(
        "log_group_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "log_uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "asset_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_log_group_version")
        ),
    )
    op.create_index(
        op.f("ix_log_group_version_end_transaction_id"),
        "log_group_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_group_version_operation_type"),
        "log_group_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_log_group_version_pk_transaction_id",
        "log_group_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_log_group_version_pk_validity",
        "log_group_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_group_version_transaction_id"),
        "log_group_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # log_group
    op.drop_index(
        op.f("ix_log_group_version_transaction_id"), table_name="log_group_version"
    )
    op.drop_index("ix_log_group_version_pk_validity", table_name="log_group_version")
    op.drop_index(
        "ix_log_group_version_pk_transaction_id", table_name="log_group_version"
    )
    op.drop_index(
        op.f("ix_log_group_version_operation_type"), table_name="log_group_version"
    )
    op.drop_index(
        op.f("ix_log_group_version_end_transaction_id"), table_name="log_group_version"
    )
    op.drop_table("log_group_version")
    op.drop_table("log_group")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
View for farmOS asset types
"""

from wuttafarm.web.views.farmos import FarmOSMasterView


class AssetTypeView(FarmOSMasterView):
    """
    View for farmOS asset types
    """

    model_name = "farmos_asset_type"
    model_title = "farmOS Asset Type"
    model_title_plural = "farmOS Asset Types"

    route_prefix = "farmos_asset_types"
    url_prefix = "/farmOS/asset-types"

    grid_columns = [
        "label",
        "description",
    ]

    sort_defaults = "label"

    form_fields = [
        "label",
        "description",
    ]

    def get_grid_data(self, columns=None, session=None):
        asset_types = self.farmos_client.resource.get("asset_type", "asset_type")
        return [self.normalize_asset_type(t) for t in asset_types["data"]]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # label
        g.set_link("label")
        g.set_searchable("label")

        # description
        g.set_searchable("description")

    def get_instance(self):
        asset_type = self.farmos_client.resource.get_id(
            "asset_type", "asset_type", self.request.matchdict["uuid"]
        )
        self.raw_json = asset_type
        return self.normalize_asset_type(asset_type["data"])

    def get_instance_title(self, asset_type):
        return asset_type["label"]

    def normalize_asset_type(self, asset_type):
        return {
            "uuid": asset_type["id"],
            "drupal_id": asset_type["attributes"]["drupal_internal__id"],
            "label": asset_type["attributes"]["label"],
            "description": asset_type["attributes"]["description"],
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # description
        f.set_widget("description", "notes")

    def get_xref_buttons(self, asset_type):
        model = self.app.model
        session = self.Session()
        buttons = []

        if wf_asset_type := (
            session.query(model.AssetType)
            .filter(model.AssetType.farmos_uuid == asset_type["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "asset_types.view", uuid=wf_asset_type.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    AssetTypeView = kwargs.get("AssetTypeView", base["AssetTypeView"])
    AssetTypeView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model definition for Units
"""

import sqlalchemy as sa

from wuttjamaican.db import model


class Measure(model.Base):
    """
    Represents a "measure" option (for quantities) from farmOS
    """

    __tablename__ = "measure"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Measure",
        "model_title_plural": "Measures",
    }

    uuid = model.uuid_column()

    ordinal = sa.Column(
        sa.Integer(),
        nullable=True,
        doc="""
        Ordinal (sequence number) for the measure.
        """,
    )

    name = sa.Column(
        sa.String(length=100),
        nullable=False,
        unique=True,
        doc="""
        Name of the measure.
        """,
    )

    drupal_id = sa.Column(
        sa.String(length=20),
        nullable=True,
        unique=True,
        doc="""
        Drupal internal ID for the measure.
        """,
    )

    def __str__(self):
        return self.name or ""


class Unit(model.Base):
    """
    Represents an "unit" (taxonomy term) from farmOS
    """

    __tablename__ = "unit"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Unit",
        "model_title_plural": "Units",
    }

    uuid = model.uuid_column()

    name = sa.Column(
        sa.String(length=100),
        nullable=False,
        unique=True,
        doc="""
        Name of the unit.
        """,
    )

    description = sa.Column(
        sa.String(length=255),
        nullable=True,
        doc="""
        Optional description for the unit.
        """,
    )

    farmos_uuid = sa.Column(
        model.UUID(),
        nullable=True,
        unique=True,
        doc="""
        UUID for the unit within farmOS.
        """,
    )

    drupal_id = sa.Column(
        sa.Integer(),
        nullable=True,
        unique=True,
        doc="""
        Drupal internal ID for the unit.
        """,
    )

    def __str__(self):
        return self.name or ""
